# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""
This module provides global definitions and functions to manage dark and flat fields
especially for tomo experiments and workflows
"""

__authors__ = ["C. Nemoz", "H.Payno"]
__license__ = "MIT"
__date__ = "06/09/2017"

import os
import re
from glob import glob
import fabio
import numpy
from queue import Queue
from tomwer.utils import docstring
from silx.io.utils import get_data
from tomwer.core import settings
from tomwer.core import utils
from tomwer.core.process.task import Task
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.utils import getDARK_N
import logging
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from . import params as dkrf_reconsparams
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.core.scan.scanfactory import ScanFactory
from tomoscan.esrf.utils import get_compacted_dataslices
from processview.core.manager import ProcessManager
from tomoscan.io import HDF5File
from processview.core.manager import DatasetState
import tomwer.version
from silx.io.utils import h5py_read_dataset
from silx.io.url import DataUrl
import typing

logger = logging.getLogger(__name__)

try:
    from tomwer.synctools.rsyncmanager import RSyncManager
except ImportError:
    logger.warning("rsyncmanager not available")
    has_rsync = True
else:
    has_rsync = False


class DarkRefs(
    Task, SuperviseProcess, Queue, input_names=("data",), output_names=("data",)
):
    """Compute median/mean dark and ref from originals (dark and ref files)"""

    WHAT_REF = "refs"
    WHAT_DARK = "dark"

    VALID_WHAT = (WHAT_REF, WHAT_DARK)
    """Tuple of valid option for What"""

    info_suffix = ".info"

    TOMO_N = "TOMO_N"

    def __init__(self, process_id=None, varinfo=None, inputs=None, label=None):
        """

        :param str file_ext:
        :param reconsparams: reconstruction parameters
        :type reconsparams: Union[None, ReconsParams, DKRFRP]
        """
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=label)
        SuperviseProcess.__init__(self, process_id=process_id)
        Queue.__init__(self)

        if inputs is None:
            inputs = {}
        if "recons_params" in inputs or "reconsparams" in inputs:
            raise ValueError(
                "wrong key: recons_params / reconsparams use dark_ref_params instead"
            )
        self._recons_params = inputs.get("dark_ref_params", None)
        if self._recons_params is None:
            self._recons_params = dkrf_reconsparams.DKRFRP()
        elif isinstance(self._recons_params, dict):
            self._recons_params = dkrf_reconsparams.DKRFRP.from_dict(
                self._recons_params
            )
        if not isinstance(self._recons_params, dkrf_reconsparams.DKRFRP):
            raise TypeError(
                f"'reconsparams' should be an instance of {dkrf_reconsparams.DKRFRP} or {dict}. Not {type(self._recons_params)}"
            )
        self._file_ext = inputs.get("file_ext", ".edf")
        if not type(self._file_ext) is str:
            raise TypeError("'file_ext' is expected to be a string")

        self._forceSync = inputs.get("force_sync", False)
        self.__new_hdf5_entry_created = False
        "used to know if the process has generated a new entry or not"

    @property
    def recons_params(self):
        return self._recons_params

    def set_recons_params(self, recons_params):
        if isinstance(recons_params, ReconsParams):
            self._recons_params = recons_params.dkrf
        elif isinstance(recons_params, dkrf_reconsparams.DKRFRP):
            self._recons_params = recons_params
        else:
            raise TypeError(
                "recons_params should be an instance of " "ReconsParams or DKRFRP"
            )

    def setPatternRecons(self, pattern):
        self._patternReconsFile = pattern

    def _instanciateReconsParams(self):
        return ReconsParams(empty=True)

    def setForceSync(self, b):
        self._forceSync = True

    @staticmethod
    def getRefHSTFiles(directory, prefix, file_ext=".edf"):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
        """
        assert isinstance(directory, str)
        res = []
        if os.path.isdir(directory) is False:
            logger.error(
                directory + " is not a directory. Cannot extract " "RefHST files"
            )
            return res

        for file in os.listdir(directory):
            if file.startswith(prefix) and file.endswith(file_ext):
                res.append(os.path.join(directory, file))
                assert os.path.isfile(res[-1])
        return res

    @staticmethod
    def getDarkHSTFiles(directory, prefix, file_ext=".edf"):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
        """
        res = []
        if os.path.isdir(directory) is False:
            logger.error(
                directory + " is not a directory. Cannot extract " "DarkHST files"
            )
            return res
        for file in os.listdir(directory):
            _prefix = prefix
            if prefix.endswith(file_ext):
                _prefix = prefix.rstrip(file_ext)
            if file.startswith(_prefix) and file.endswith(file_ext):
                _file = file.lstrip(_prefix).rstrip(file_ext)
                if _file == "" or _file.isnumeric() is True:
                    res.append(os.path.join(directory, file))
                    assert os.path.isfile(res[-1])
        return res

    @staticmethod
    def getDarkPatternTooltip():
        return (
            "define the pattern to find, using the python `re` library.\n"
            "For example: \n"
            "   - `.*conti_dark.*` to filter files containing `conti_dark` sentence\n"
            "   - `darkend[0-9]{3,4}` to filter files named `darkend` followed by three or four digit characters (and having the .edf extension)"
        )

    @staticmethod
    def getRefPatternTooltip():
        return (
            "define the pattern to find, using the python `re` library.\n"
            "For example: \n"
            "   - `.*conti_ref.*` for files containing `conti_dark` sentence\n"
            "   - `ref*.*[0-9]{3,4}_[0-9]{3,4}` to filter files named `ref` followed by any character and ending by X_Y where X and Y are groups of three or four digit characters."
        )

    @staticmethod
    def properties_help():
        return """
        - refs: 'None', 'Median', 'Average', 'First', 'Last' \n
        - dark: 'None', 'Median', 'Average', 'First', 'Last' \n
        """

    def set_configuration(self, properties):
        # No properties stored for now
        if "dark" in properties:
            self._recons_params.dark_calc_method = properties["dark"]
        if "refs" in properties:
            self._recons_params.ref_calc_method = properties["refs"]
        if "_rpSetting" in properties:
            self._recons_params.load_from_dict(properties["_rpSetting"])
        else:
            self._recons_params.load_from_dict(properties)

    @staticmethod
    def get_darks_frm_process_file(
        process_file, entry=None, as_url: bool = False
    ) -> typing.Union[None, dict]:
        """

        :param str process_file: path to the process file
        :param entry: entry to read in the process file if more than one
        :param bool as_url: if true then an url will be used instead of a
                            numpy.array
        :return: dictionary with index in the sequence as key and numpy array
                 as value (or url if as_url set to True)
        """
        if entry is None:
            with HDF5File(process_file, "r", swmr=True) as h5f:
                entries = DarkRefs._get_process_nodes(root_node=h5f, process=DarkRefs)
                if len(entries) == 0:
                    logger.info("unable to find a DarkRef process in %s" % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    logger.info("take %s as default entry" % entry)

        with HDF5File(process_file, "r", swmr=True) as h5f:
            if entry not in h5f.keys():
                logger.info("no dark found for {}".format(entry))
                return {}
            dark_nodes = DarkRefs._get_process_nodes(
                root_node=h5f[entry], process=DarkRefs
            )
            index_to_path = {}
            for key, index in dark_nodes.items():
                index_to_path[index] = key

            if len(dark_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(list(dark_nodes.values()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                logger.debug(
                    "several processing found for dark-ref,"
                    "take the last one: %s" % last_process_dark
                )

            res = {}
            if "results" in h5f[last_process_dark].keys():
                results_node = h5f[last_process_dark]["results"]
                if "darks" in results_node.keys():
                    darks = results_node["darks"]
                    for index in darks:
                        if as_url is True:
                            res[int(index)] = DataUrl(
                                file_path=process_file,
                                data_path="/".join((darks.name, index)),
                                scheme="silx",
                            )
                        else:
                            res[int(index)] = h5py_read_dataset(darks[index])
            return res

    @staticmethod
    def get_flats_frm_process_file(
        process_file, entry=None, as_url: bool = False
    ) -> typing.Union[None, dict]:
        """

        :param process_file:
        :param entry: entry to read in the process file if more than one
        :param bool as_url: if true then an url will be used instead of a
                            numpy.array
        :return:
        """
        if entry is None:
            with HDF5File(process_file, "r", swmr=True) as h5f:
                entries = DarkRefs._get_process_nodes(root_node=h5f, process=DarkRefs)
                if len(entries) == 0:
                    logger.info("unable to find a DarkRef process in %s" % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    logger.info("take %s as default entry" % entry)

        with HDF5File(process_file, "r", swmr=True) as h5f:
            if entry not in h5f.keys():
                logger.info("no flats found for {}".format(entry))
                return {}
            dkref_nodes = DarkRefs._get_process_nodes(
                root_node=h5f[entry], process=DarkRefs
            )
            index_to_path = {}
            for key, index in dkref_nodes.items():
                index_to_path[index] = key

            if len(dkref_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dkref_nodes.values())[-1]
            last_process_dkrf = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                logger.debug(
                    "several processing found for dark-ref,"
                    "take the last one: %s" % last_process_dkrf
                )

            res = {}
            if "results" in h5f[last_process_dkrf].keys():
                results_node = h5f[last_process_dkrf]["results"]
                if "flats" in results_node.keys():
                    flats = results_node["flats"]
                    for index in flats:
                        if as_url is True:
                            res[int(index)] = DataUrl(
                                file_path=process_file,
                                data_path="/".join((flats.name, index)),
                                scheme="silx",
                            )
                        else:
                            res[int(index)] = h5py_read_dataset(flats[index])
            return res

    def run(self):
        scan = self.inputs.data
        if scan is None:
            self.outputs.data = None
            return

        if type(scan) is str:
            assert os.path.exists(scan)
            scan = ScanFactory.create_scan_object(scan_path=scan)
        elif isinstance(scan, TomwerScanBase):
            pass
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise TypeError(
                "scan should be an instance of TomoBase or path to " "scan dircetory"
            )
        # copy the current parameters to the scan ftseries_recons_params
        if scan.ftseries_recons_params is None:
            scan.ftseries_recons_params = self._instanciateReconsParams()
        assert isinstance(self._recons_params, dkrf_reconsparams.DKRFRP)
        assert isinstance(scan.ftseries_recons_params, ReconsParams)
        assert self._recons_params is not None

        scan.ftseries_recons_params.copy(self._recons_params)

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=DatasetState.ON_GOING,
        )
        logger.processStarted("start dark and ref for {}" "".format(str(scan)))
        if (
            settings.isOnLbsram(scan)
            and utils.isLowOnMemory(settings.get_lbsram_path()) is True
        ):
            mess = (
                "low memory, do compute dark and flat field mean/median "
                "for %s" % scan.path
            )
            logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=DatasetState.SKIPPED,
                details=mess,
            )
            self.outputs.data = None
            return

        if not (scan and os.path.exists(scan.path)):
            mess = "folder {} is not existing".format(scan.path)
            logger.warning(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan, process=self, state=DatasetState.FAILED, details=mess
            )

            self.outputs.data = None
            return
        whats = (DarkRefs.WHAT_REF, DarkRefs.WHAT_DARK)
        modes = (
            self.recons_params.ref_calc_method,
            self.recons_params.dark_calc_method,
        )

        for what, mode in zip(whats, modes):
            self._originalsDark = []
            self._originalsRef = []
            logger.debug(
                "compute {what} using mode {mode} for {scan}"
                "".format(what=what, mode=mode, scan=str(scan))
            )
            self.compute(scan=scan, what=what, mode=mode)
        try:
            for what, mode in zip(whats, modes):
                self._originalsDark = []
                self._originalsRef = []
                self.compute(scan=scan, what=what, mode=mode)
        except Exception as e:
            info = "Fail computing dark and ref for {}. Reason is {}".format(
                str(scan), e
            )
            self.process.notify_to_state_to_managed(
                dataset=scan, state=DatasetState.SUCCEED, details=info
            )
            logger.processFailed(info)
            self.outputs.data = None
            return
        results = {}
        interpretations = {}
        if (
            self.recons_params.dark_calc_method is not dkrf_reconsparams.Method.none
            and scan.normed_darks is not None
        ):
            # cast darks and flats keys from int (index) to str
            o_darks = scan.normed_darks
            f_darks = {}
            for index, data in o_darks.items():
                f_darks[str(index)] = data
                interpretations["/".join(("darks", str(index)))] = "image"
            results["darks"] = f_darks
            # EDFTomoscan are already saved
            if not isinstance(scan, EDFTomoScan):
                scan.save_reduced_darks(f_darks)
        if (
            self.recons_params.ref_calc_method is not dkrf_reconsparams.Method.none
            and scan.normed_flats is not None
        ):
            results["flats"] = scan.normed_flats
            o_flats = scan.normed_flats
            f_flats = {}
            for index, data in o_flats.items():
                f_flats[str(index)] = data
                interpretations["/".join(("flats", str(index)))] = "image"
            results["flats"] = f_flats
            # EDFTomoscan are already saved
            if not isinstance(scan, EDFTomoScan):
                scan.save_reduced_flats(f_flats)

        if len(results) > 0:
            # if some processing to be registered
            if scan.process_file is not None and not (
                isinstance(scan, HDF5TomoScan) and not self.__new_hdf5_entry_created
            ):
                with scan.acquire_process_file_lock():
                    entry = "entry"
                    if hasattr(scan, "entry"):
                        entry = scan.entry
                    self.register_process(
                        process_file=scan.process_file,
                        entry=entry,
                        configuration=self.recons_params.to_dict(),
                        results=results,
                        interpretations=interpretations,
                        process_index=scan.pop_process_index(),
                        overwrite=True,
                    )
        logger.processSucceed("Dark and ref succeeded for {}" "".format(str(scan)))
        self.notify_to_state_to_managed(
            dataset=scan, state=DatasetState.SUCCEED, details=None
        )

        if self._return_dict:
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    def compute(self, scan, what, mode):
        if isinstance(scan, EDFTomoScan):
            self.compute_edf(scan, what=what, mode=mode)
        elif isinstance(scan, HDF5TomoScan):
            self.compute_hdf5(scan, what=what, mode=mode)
        else:
            raise ValueError("scan type is not recognized ofr %s" % scan)

    def compute_hdf5(self, scan, what, mode):
        """Compute the requested what in the given mode for `directory`

        :param str directory: path of the scan
        :param what: what to compute (ref or dark)
        :param mode: how to compute it (median or average...)
        """

        def get_series(scan, what: str) -> list:
            """return a list of dictionary. on the dictionary we have indexes
            as key and url as value"""
            if what == "dark":
                raw_what = scan.darks
            else:
                raw_what = scan.flats
            if len(raw_what) == 0:
                return []
            else:
                series = []
                indexes = sorted(raw_what.keys())
                # a serie is defined by contiguous indexes
                current_serie = {indexes[0]: raw_what[indexes[0]]}
                current_index = indexes[0]
                for index in indexes[1:]:
                    if index == current_index + 1:
                        current_index = index
                    else:
                        series.append(current_serie)
                        current_serie = {}
                        current_index = index
                    current_serie[index] = raw_what[index]
                if len(current_serie) > 0:
                    series.append(current_serie)
                return series

        if mode is dkrf_reconsparams.Method.median:
            method_ = numpy.median
        elif mode is dkrf_reconsparams.Method.average:
            method_ = numpy.mean
        elif mode is dkrf_reconsparams.Method.none:
            return None
        elif mode is dkrf_reconsparams.Method.first:
            method_ = "raw"
        elif mode is dkrf_reconsparams.Method.last:
            method_ = "raw"
        else:
            raise ValueError(
                "Mode {mode} for {what} is not managed" "".format(mode=mode, what=what)
            )
        raw_series = get_series(scan, what)
        if len(raw_series) == 0:
            logger.warning("No %s found for %s" % (what, scan))
            return

        def load_data_serie(urls):
            if mode is dkrf_reconsparams.Method.first and len(urls) > 0:
                urls_keys = sorted(urls.keys())
                urls = {
                    urls_keys[0]: urls[urls_keys[0]],
                }
            if mode is dkrf_reconsparams.Method.last and len(urls) > 0:
                urls_keys = sorted(urls.keys())
                urls = {
                    urls_keys[-1]: urls[urls_keys[-1]],
                }

            cpt_slices = get_compacted_dataslices(urls)
            url_set = {}
            for url in cpt_slices.values():
                path = url.file_path(), url.data_path(), str(url.data_slice())
                url_set[path] = url

            n_elmts = 0
            for url in url_set.values():
                my_slice = url.data_slice()
                n_elmts += my_slice.stop - my_slice.start

            data = None
            start_z = 0
            for url in url_set.values():
                my_slice = url.data_slice()
                my_slice = slice(my_slice.start, my_slice.stop, 1)
                new_url = DataUrl(
                    file_path=url.file_path(),
                    data_path=url.data_path(),
                    data_slice=my_slice,
                    scheme="silx",
                )
                loaded_data = get_data(new_url)

                # init data if dim is not know
                if data is None:
                    data = numpy.empty(
                        shape=(
                            n_elmts,
                            scan.dim_2 or loaded_data.shape[-2],
                            scan.dim_1 or loaded_data.shape[-1],
                        )
                    )
                if loaded_data.ndim == 2:
                    data[start_z, :, :] = loaded_data
                    start_z += 1
                elif loaded_data.ndim == 3:
                    delta_z = my_slice.stop - my_slice.start
                    data[start_z:delta_z, :, :] = loaded_data
                    start_z += delta_z
                else:
                    raise ValueError("Dark and ref raw data should be 2D or 3D")

            return data

        res = {}
        # res: index: sequence when the serie was taken

        self.__new_hdf5_entry_created = False
        # flag to know if we could load all dark and flats from a previous
        # process file or if we add to create a new entry
        for serie_ in raw_series:
            serie_index = min(serie_)
            if what == "dark" and len(res) > 0:
                continue
            old_data = None
            has_p_file = os.path.exists(scan.process_file)
            if (
                has_p_file
                and what == DarkRefs.WHAT_DARK
                and not self.recons_params.overwrite_dark
            ):
                old_data = DarkRefs.get_darks_frm_process_file(
                    scan.process_file, entry=scan.entry
                )
            elif (
                has_p_file
                and what == DarkRefs.WHAT_REF
                and not self.recons_params.overwrite_ref
            ):
                old_data = DarkRefs.get_flats_frm_process_file(
                    scan.process_file, entry=scan.entry
                )
            if old_data is not None and serie_index in old_data:
                logger.info("load {} from existing data".format(what))
                res[serie_index] = old_data[serie_index]
                continue
            self.__new_hdf5_entry_created = True
            serie_data = load_data_serie(serie_)
            if method_ == "raw":
                res[serie_index] = serie_data.reshape(-1, serie_data.shape[-1])
            else:
                res[serie_index] = method_(serie_data, axis=0)
        if what == "dark":
            scan.set_normed_darks(res)
        else:
            scan.set_normed_flats(res)

    def compute_edf(self, scan, what, mode):
        """Compute the requested what in the given mode for `directory`

        :param str directory: path of the scan
        :param what: what to compute (ref or dark)
        :param mode: how to compute it (median or average...)
        """
        directory = scan.path
        assert type(directory) is str

        def removeFiles():
            """Remove orignals files fitting the what (dark or ref)"""
            if what is DarkRefs.WHAT_DARK:
                # In the case originals has already been found for the median
                # calculation
                if len(self._originalsDark) > 0:
                    files = self._originalsDark
                else:
                    files = getOriginals(DarkRefs.WHAT_DARK)
            elif what is DarkRefs.WHAT_REF:
                if len(self._originalsRef) > 0:
                    files = self._originalsRef
                else:
                    files = getOriginals(DarkRefs.WHAT_REF)
            else:
                logger.error(
                    "the requested what (%s) is not recognized. "
                    "Can't remove corresponding file" % what
                )
                return

            _files = set(files)
            if len(files) > 0:
                if has_rsync:
                    logger.info(
                        "ask RSyncManager for removal of %s files in %s"
                        % (what, directory)
                    )
                    # for lbsram take into account sync from data watcher
                    if directory.startswith(settings.get_lbsram_path()):
                        for f in files:
                            _files.add(
                                f.replace(
                                    settings.get_lbsram_path(),
                                    settings.get_dest_path(),
                                    1,
                                )
                            )
                    RSyncManager().removesync_files(
                        dir=directory, files=_files, block=self._forceSync
                    )
                else:
                    for _file in _files:
                        try:
                            os.remove(_file)
                        except Exception as e:
                            logger.error(e)

        def getOriginals(what):
            if what is DarkRefs.WHAT_REF:
                try:
                    pattern = re.compile(self.recons_params.ref_pattern)
                except Exception:
                    pattern = None
                    logger.error(
                        "Fail to compute regular expresion for %s"
                        % self.recons_params.ref_pattern
                    )
            elif what is DarkRefs.WHAT_DARK:
                re.compile(self.recons_params.dark_pattern)
                try:
                    pattern = re.compile(self.recons_params.dark_pattern)
                except Exception:
                    pattern = None
                    logger.error(
                        "Fail to compute regular expresion for %s"
                        % self.recons_params.dark_pattern
                    )

            filelist_fullname = []
            if pattern is None:
                return filelist_fullname
            for file in os.listdir(directory):
                if pattern.match(file) and file.endswith(self._file_ext):
                    if (
                        file.startswith(self.recons_params.ref_prefix)
                        or file.startswith(self.recons_params.dark_prefix)
                    ) is False:
                        filelist_fullname.append(os.path.join(directory, file))
            return sorted(filelist_fullname)

        def setup():
            """setup parameter to process the requested what

            :return: True if there is a process to be run, else false
            """

            def getNDigits(_file):
                file_without_scanID = _file.replace(os.path.basename(directory), "", 1)
                return len(re.findall(r"\d+", file_without_scanID))

            def dealWithPCOTomo():
                filesPerSerie = {}
                if self.nfiles % self.nacq == 0:
                    assert self.nacq < self.nfiles
                    self.nseries = self.nfiles // self.nacq
                    self.series = self.fileNameList
                else:
                    logger.warning("Fail to deduce series")
                    return None, None

                linear = getNDigits(self.fileNameList[0]) < 2
                if linear is False:
                    # which digit pattern contains the file number?
                    lastone = True
                    penulti = True
                    for first_files in range(self.nseries - 1):
                        digivec_1 = re.findall(r"\d+", self.fileNameList[first_files])
                        digivec_2 = re.findall(
                            r"\d+", self.fileNameList[first_files + 1]
                        )
                        if lastone:
                            lastone = (int(digivec_2[-1]) - int(digivec_1[-1])) == 0
                        if penulti:
                            penulti = (int(digivec_2[-2]) - int(digivec_1[-2])) == 0

                    linear = not penulti

                if linear is False:
                    digivec_1 = re.findall(r"\d+", self.fileNameList[self.nseries - 1])
                    digivec_2 = re.findall(r"\d+", self.fileNameList[self.nseries])
                    # confirm there is 1 increment after self.nseries in the uperlast last digit patern
                    if (int(digivec_2[-2]) - int(digivec_1[-2])) != 1:
                        linear = True

                # series are simple sublists in main filelist
                # self.series = []
                if linear is True:
                    is_there_digits = len(re.findall(r"\d+", self.fileNameList[0])) > 0
                    if is_there_digits:
                        serievec = set([re.findall(r"\d+", self.fileNameList[0])[-1]])
                    else:
                        serievec = set(["0000"])
                    for i in range(self.nseries):
                        if is_there_digits:
                            serie = re.findall(
                                r"\d+", self.fileNameList[i * self.nacq]
                            )[-1]
                            serievec.add(serie)
                            filesPerSerie[serie] = self.fileNameList[
                                i * self.nacq : (i + 1) * self.nacq
                            ]
                        else:
                            serievec.add("%04d" % i)
                # in the sorted filelist, the serie is incremented, then the acquisition number:
                else:
                    self.series = self.fileNameList[0 :: self.nseries]
                    serievec = set([re.findall(r"\d+", self.fileNameList[0])[-1]])
                    for serie in serievec:
                        # serie = re.findall(r'\d+', self.fileNameList[i])[-1]
                        # serievec.add(serie)
                        filesPerSerie[serie] = self.fileNameList[0 :: self.nseries]
                serievec = list(sorted(serievec))

                if len(serievec) > 2:
                    logger.error(
                        "DarkRefs do not deal with multiple scan."
                        " (scan %s)" % directory
                    )
                    return None, None
                assert len(serievec) <= 2
                if len(serievec) > 1:
                    key = serievec[-1]
                    tomoN = self.getInfo(self.TOMO_N)
                    if tomoN is None:
                        logger.error(
                            "Fail to found information %s. Can't "
                            "rename %s" % (self.TOMO_N, key)
                        )
                    del serievec[-1]
                    serievec.append(str(tomoN).zfill(4))
                    filesPerSerie[serievec[-1]] = filesPerSerie[key]
                    del filesPerSerie[key]
                    assert len(serievec) == 2
                    assert len(filesPerSerie) == 2

                return serievec, filesPerSerie

            # start setup function
            if mode == dkrf_reconsparams.Method.none:
                return False
            if what == "dark":
                self.out_prefix = self.recons_params.dark_prefix
                self.info_nacq = "DARK_N"
            else:
                self.out_prefix = self.recons_params.ref_prefix
                self.info_nacq = "REF_N"

            # init
            self.nacq = 0
            """Number of acquisition runned"""
            self.files = 0
            """Ref or dark files"""
            self.nframes = 1
            """Number of frame per ref/dark file"""
            self.serievec = ["0000"]
            """List of series discover"""
            self.filesPerSerie = {}
            """Dict with key the serie id and values list of files to compute
            for median or mean"""
            self.infofile = ""
            """info file of the acquisition"""

            # sample/prefix and info file
            self.prefix = os.path.basename(directory)
            extensionToTry = (DarkRefs.info_suffix, "0000" + DarkRefs.info_suffix)
            for extension in extensionToTry:
                infoFile = os.path.join(directory, self.prefix + extension)
                if os.path.exists(infoFile):
                    self.infofile = infoFile
                    break

            if self.infofile == "":
                logger.debug("fail to found .info file for %s" % directory)

            """
            Set filelist
            """
            # do the job only if not already done and overwrite not asked
            self.out_files = sorted(glob(directory + os.sep + "*." + self._file_ext))

            self.filelist_fullname = getOriginals(what)
            self.fileNameList = []
            [
                self.fileNameList.append(os.path.basename(_file))
                for _file in self.filelist_fullname
            ]
            self.fileNameList = sorted(self.fileNameList)
            self.nfiles = len(self.filelist_fullname)
            # if nothing to process
            if self.nfiles == 0:
                logger.info(
                    "no %s for %s, because no file to compute found" % (what, directory)
                )
                return False

            self.fid = fabio.open(self.filelist_fullname[0])
            self.nframes = self.fid.nframes
            self.nacq = 0
            # get the info of number of acquisitions
            if self.infofile != "":
                self.nacq = self.getInfo(self.info_nacq)

            if self.nacq == 0:
                self.nacq = self.nfiles

            self.nseries = 1
            if self.nacq > self.nfiles:
                # get ready for accumulation and/or file multiimage?
                self.nseries = self.nfiles
            if self.nacq < self.nfiles and getNDigits(self.fileNameList[0]) < 2:
                self.nFilePerSerie = self.nseries
                self.serievec, self.filesPerSerie = dealWithPCOTomo()
            else:
                self.series = self.fileNameList
                self.serievec = _getSeriesValue(self.fileNameList)
                self.filesPerSerie, self.nFilePerSerie = groupFilesPerSerie(
                    self.filelist_fullname, self.serievec
                )

            if self.filesPerSerie is not None:
                for serie in self.filesPerSerie:
                    for _file in self.filesPerSerie[serie]:
                        if what == "dark":
                            self._originalsDark.append(os.path.join(scan.path, _file))
                        elif what == "ref":
                            self._originalsRef.append(os.path.join(scan.path, _file))

            return self.serievec is not None and self.filesPerSerie is not None

        def _getSeriesValue(fileNames):
            assert len(fileNames) > 0
            is_there_digits = len(re.findall(r"\d+", fileNames[0])) > 0
            series = set()
            i = 0
            for fileName in fileNames:
                if is_there_digits:
                    name = fileName.rstrip(self._file_ext)
                    file_index = name.split("_")[-1]
                    rm_not_numeric = re.compile(r"[^\d.]+")
                    file_index = rm_not_numeric.sub("", file_index)
                    series.add(file_index)
                else:
                    series.add("%04d" % i)
                    i += 1
            return list(series)

        def groupFilesPerSerie(files, series):
            def findFileEndingWithSerie(poolFiles, serie):
                res = []
                for _file in poolFiles:
                    _f = _file.rstrip(".edf")
                    if _f.endswith(serie):
                        res.append(_file)
                return res

            def checkSeriesFilesLength(serieFiles):
                length = -1
                for serie in serieFiles:
                    if length == -1:
                        length = len(serieFiles[serie])
                    elif len(serieFiles[serie]) != length:
                        logger.error("Series with inconsistant number of ref files")

            assert len(series) > 0
            if len(series) == 1:
                return {series[0]: files}, len(files)
            assert len(files) > 0

            serieFiles = {}
            unattributedFiles = files.copy()
            for serie in series:
                serieFiles[serie] = findFileEndingWithSerie(unattributedFiles, serie)
                [unattributedFiles.remove(_f) for _f in serieFiles[serie]]

            if len(unattributedFiles) > 0:
                logger.error("Failed to associate %s to any serie" % unattributedFiles)
                return {}, 0

            checkSeriesFilesLength(serieFiles)

            return serieFiles, len(serieFiles[list(serieFiles.keys())[0]])

        def process():
            """process calculation of the what"""
            if mode is dkrf_reconsparams.Method.none:
                return
            shape = fabio.open(self.filelist_fullname[0]).shape

            for i in range(len(self.serievec)):
                largeMat = numpy.zeros(
                    (self.nframes * self.nFilePerSerie, shape[0], shape[1])
                )

                if what == "dark" and len(self.serievec) == 1:
                    fileName = self.out_prefix
                    if fileName.endswith(self._file_ext) is False:
                        fileName = fileName + self._file_ext
                else:
                    fileName = (
                        self.out_prefix.rstrip(self._file_ext)
                        + self.serievec[i]
                        + self._file_ext
                    )
                fileName = os.path.join(directory, fileName)
                if os.path.isfile(fileName):
                    if (
                        what == "refs" and self.recons_params.overwrite_ref is False
                    ) or (
                        what == "dark" and self.recons_params.overwrite_dark is False
                    ):
                        logger.info("skip creation of %s, already existing" % fileName)
                        continue

                if self.nFilePerSerie == 1:
                    fSerieName = os.path.join(directory, self.series[i])
                    header = {"method": mode.name + " on 1 image"}
                    header["SRCUR"] = utils.getClosestSRCurrent(
                        scan_dir=directory, refFile=fSerieName
                    )
                    if self.nframes == 1:
                        largeMat[0] = fabio.open(fSerieName).data
                    else:
                        handler = fabio.open(fSerieName)
                        dShape = (self.nframes, handler.dim2, handler.dim1)
                        largeMat = numpy.zeros(dShape)
                        for iFrame in range(self.nframes):
                            largeMat[iFrame] = handler.getframe(iFrame).data
                else:
                    header = {
                        "method": mode.name + " on %d images" % self.nFilePerSerie
                    }
                    header["SRCUR"] = utils.getClosestSRCurrent(
                        scan_dir=directory, refFile=self.series[i][0]
                    )
                    for j, fName in zip(
                        range(self.nFilePerSerie), self.filesPerSerie[self.serievec[i]]
                    ):
                        file_BigMat = fabio.open(fName)
                        if self.nframes > 1:
                            for fr in range(self.nframes):
                                jfr = fr + j * self.nframes
                                largeMat[jfr] = file_BigMat.getframe(fr).getData()
                        else:
                            largeMat[j] = file_BigMat.data
                if mode == dkrf_reconsparams.Method.median:
                    data = numpy.median(largeMat, axis=0)
                elif mode == dkrf_reconsparams.Method.average:
                    data = numpy.mean(largeMat, axis=0)
                elif mode == dkrf_reconsparams.Method.first:
                    data = largeMat[0]
                elif mode == dkrf_reconsparams.Method.last:
                    data = largeMat[-1]
                elif mode == dkrf_reconsparams.Method.none:
                    return
                else:
                    raise ValueError(
                        "Unrecognized calculation type request {}" "".format(mode)
                    )

                self.nacq = getDARK_N(directory) or 1
                if what == "dark" and self.nacq > 1:  # and self.nframes == 1:
                    data = data / self.nacq
                    # add one to add to avoid division by zero
                    # data = data + 1
                file_desc = fabio.edfimage.EdfImage(data=data, header=header)
                i += 1
                _ttype = numpy.uint16 if what == "dark" else numpy.int32
                file_desc.write(fileName, force_type=_ttype)

        if directory is None:
            return
        if setup():
            logger.info("start proccess darks and flat fields for %s" % scan.path)
            process()
            logger.info("end proccess darks and flat fields")

        self._store_result(what=what, scan=scan)
        if (what == "dark" and self.recons_params.remove_dark is True) or (
            what == "refs" and self.recons_params.remove_ref is True
        ):
            removeFiles()

    def _store_result(self, what, scan):
        try:
            if what == "dark":
                dark = DarkRefs.getDarkHSTFiles(
                    directory=scan.path, prefix=self.recons_params.dark_prefix
                )
                dark_1 = fabio.open(dark[0]).data
                scan.set_normed_darks({0: dark_1})
            else:
                refs = DarkRefs.getRefHSTFiles(
                    directory=scan.path, prefix=self.recons_params.ref_prefix
                )
                refs_dict = {}
                if len(refs) == 1:
                    ref_1 = fabio.open(refs[0]).data
                    refs_dict[0] = ref_1
                elif len(refs) >= 2:
                    ref_1 = fabio.open(refs[0]).data
                    refs_dict[0] = ref_1
                    ref_f = fabio.open(refs[-1]).data
                    refs_dict[len(scan.projections)] = ref_f
                scan.set_normed_flats(refs_dict)
        except Exception as e:
            logger.info(e)

    def getInfo(self, what):
        with open(self.infofile) as file:
            infod = file.readlines()
            for line in infod:
                if what in line:
                    return int(line.split("=")[1])
        # not found:
        return 0

    def getDarkFiles(self, directory):
        """

        :return: the list of existing darks files in the directory according to
                 the file pattern.
        """
        patternDark = re.compile(self.recons_params.dark_pattern)

        res = []
        for file in os.listdir(directory):
            if patternDark.match(file) is not None and file.endswith(self._file_ext):
                res.append(os.path.join(directory, file))
        return res

    def getRefFiles(self, directory):
        """

        :return: the list of existing refs files in the directory according to
                 the file pattern.
        """
        patternRef = re.compile(self.recons_params.ref_pattern)

        res = []
        for file in os.listdir(directory):
            if patternRef.match(file) and file.endswith(self._file_ext):
                res.append(os.path.join(directory, file))
        return res

    @docstring(Task.program_name)
    @staticmethod
    def program_name():
        return "tomwer_dark_refs"

    @docstring(Task.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(Task.definition)
    @staticmethod
    def definition():
        return "Compute mean or median dark and refs per each serie"


def requires_normed_dark_and_flat(scan: TomwerScanBase, logger_=None) -> tuple:
    """helper function: If no dark / flat are computed yet then will pick the first
    dark and the first flat.

    Expected usage: for tomwer application which requires some time and flat and to avoid some warnings
    within standalones.

    :params TomwerScanBase scan: scan for which we want to get quick dark and flat
    :params Optional[Logger] logger_: if provided will add some warning when attempt to get normed flat / normed dark.
    :returns: tuple of what was missing and has been computed
    """
    computed = []
    if scan.normed_flats in (None, {}):
        # set the first flat found
        recons_params = dkrf_reconsparams.DKRFRP()
        recons_params.overwrite_dark = False
        recons_params.overwrite_ref = False
        recons_params.dark_calc_method = dkrf_reconsparams.Method.none
        recons_params.flat_calc_method = dkrf_reconsparams.Method.first

        drp = DarkRefs(
            inputs={
                "data": scan,
                "dark_ref_params": recons_params,
            }
        )
        if logger_ is not None:
            logger.warning(
                "No 'calculated' flat found. Will try to pick the first flat found as the `calculated` flat."
            )
        drp.run()
        computed.append("flat")

    if scan.normed_darks in (None, {}):
        # set the first dark found
        recons_params = dkrf_reconsparams.DKRFRP()
        recons_params.overwrite_dark = False
        recons_params.overwrite_ref = False
        recons_params.dark_calc_method = dkrf_reconsparams.Method.first
        recons_params.flat_calc_method = dkrf_reconsparams.Method.none

        drp = DarkRefs(
            inputs={
                "data": scan,
                "dark_ref_params": recons_params,
            }
        )
        if logger_ is not None:
            logger.warning(
                "No 'calculated' dark found. Will try to pick the first flat found as the `calculated` dark."
            )
        drp.run()
        computed.append("dark")

    return tuple(computed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
import argparse
import glob
import os
import shutil
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.synctools.utils.scanstages import ScanStages


logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer rsync [source_dir] [dest_dir] [options]"


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "source_dir",
        help="folder containing the acquisition to copy."
        "To copy /dir_path/scan_dir into /dir_path2/scan_dir source-dir "
        "should be /dir_path/scan_dir",
    )
    parser.add_argument(
        "dest_dir",
        help="destination folder. If doesn't exists will create it."
        "Will synchronize"
        "If exists and contains some files, those won't be remove"
        "except if the --fresh option is activated"
        "To copy /dir_path/scan_dir into /dir_path2/scan_dir dest-dir "
        "should be /dir_path2",
    )
    parser.add_argument(
        "--advancement",
        help="An acquisition contains several steps (initialization, "
        "acquisition, complete...). You can ask to synchronize file until"
        " a specific stage. \n"
        "For edf for example if you specify the stage `on_going` no .info"
        " while be created and the targeted directory won't contain "
        "any 'slice' files. \n"
        "Default value is `complete`"
        "Valid advancement values are: %s"
        % str(ScanStages.AcquisitionStage.get_command_names()),
    )
    parser.add_argument(
        "--fresh",
        help="first file to be used for computing axis",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--force", "-f", help="force folder removal", action="store_true", default=False
    )
    parser.add_argument(
        "--rm-final-xml",
        help="util option: remove the .xml created at the end of some .edf"
        "acquisition (which notify the end of the acquisition). \n"
        "Warning: this is an exclusive option. So it cannot be associated"
        "with any other option.",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(argv[1:])

    if options.rm_final_xml is True and (
        (options.advancement is not None) or (options.fresh is True)
    ):
        logger.error(
            "--rm-final-xml is an `exclusive`option. It cannot be"
            "associated with an other option"
        )
        return

    if options.advancement is None and options.rm_final_xml is False:
        options.advancement = "complete"

    for scan_path in glob.glob(options.source_dir):
        try:
            scan = ScanFactory.create_scan_object(scan_path=scan_path)
        except ValueError:
            logger.error(scan_path + " is not containing a scan")
            continue

        dest_dir = os.path.join(options.dest_dir, os.path.basename(options.source_dir))

        # fresh option treatment
        def refresh_dir(dir_):
            logger.info("recreate", dir_)
            shutil.rmtree(dir_)
            os.makedirs(dir_)

        if options.rm_final_xml is True:
            final_xml = os.path.join(dest_dir, os.path.basename(dest_dir) + ".xml")
            if os.path.exists(final_xml):
                os.remove(final_xml)
            # if the xml option is created then other options are ignore
            continue

        if options.fresh and os.path.exists(dest_dir):
            if options.force is False:
                conf = input("do you want to remove %s (yes/y) ?" % dest_dir)
                if conf.lower() in ("yes", "y"):
                    refresh_dir(dir_=dest_dir)
            else:
                refresh_dir(dir_=dest_dir)

        scan_stages = ScanStages(scan=scan)
        stage = ScanStages.AcquisitionStage.from_command_name(options.advancement)
        scan_stages.rsync_until(stage=stage, dest_dir=dest_dir)


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
Some utils GUI associated to illustrations
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "29/01/2020"


import subprocess
import logging

_logger = logging.getLogger(__name__)


__has_image_j = None


def has_imagej():
    """Return if imagej command is accessible from the computer"""
    global __has_image_j
    if __has_image_j is None:
        try:
            # use help because there is no information regarding version
            subprocess.call(["imagej", "-h"], stdout=subprocess.PIPE)
        except OSError:
            __has_image_j = False
        else:
            __has_image_j = True
    return __has_image_j


def open_url_with_image_j(url):
    """open the url in an imagej subprocess

    :param DataUrl url: url we want to open in imagej
    """
    if not has_imagej():
        raise OSError("ImageJ is not installed")
    else:
        # for now we only manage the simple case of an edf file
        try:
            subprocess.call(["imagej", "-o", url.file_path()])
        except Exception:
            _logger.warning("Fail to open {}. Reason is {}".format(str(url), str()))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import shutil
import tempfile
import unittest
import os
import time
import pytest
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.process.control.datawatcher import DataWatcher


class TestDataWatcherIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()
        self.data_watcher_process = None
        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )

    def tearDown(self):
        if self.data_watcher_process is not None:
            self.data_watcher_process.stop()
        shutil.rmtree(self.scan_folder)

    @pytest.mark.skipif(os.name != "posix", reason="not tested under windows yet")
    def testOutputDict(self):
        self.data_watcher_process = DataWatcher()
        self.data_watcher_process.setFolderObserved(self.scan_folder)
        self.data_watcher_process.setWaitTimeBtwLoop(1)

        self.data_watcher_process.clear_output_values()
        self.data_watcher_process._set_return_dict(True)
        self.data_watcher_process.start()

        timeout = 8
        while (
            self.data_watcher_process.get_output_value("data") is None and timeout > 0
        ):
            _delta = 0.5
            time.sleep(_delta)
            timeout -= _delta

        if timeout <= 0:
            self.fail("timeout expire")
        out = self.data_watcher_process.get_output_value("data")
        self.assertTrue(isinstance(out, dict))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/03/2019"


import enum
from .base import _ReconsParam


@enum.unique
class BeamGeoType(enum.Enum):
    parallel = (0,)
    fan = (2,)
    conical = (3,)


class BeamGeoRP(_ReconsParam):
    """Beam geometry reconstruction parameters"""

    def __init__(self):
        _ReconsParam.__init__(self)
        self.__type = BeamGeoType.parallel
        self.__sx = 0.0
        self.__sy = 0.0
        self.__dist = 55.0

        self._managed_params = {
            "TYPE": self.__class__.type,
            "SX": self.__class__.sx,
            "SY": self.__class__.sy,
            "DIST": self.__class__.dist,
        }

    @property
    def type(self):
        """Beam geometry type"""
        return self.__type

    @type.setter
    def type(self, _type):
        assert isinstance(_type, (BeamGeoType, int, str))
        if type(_type) is str:
            __type = self._get_beam_geo_frm_letter(_type)
        else:
            __type = BeamGeoType(_type)
        if __type != self.__type:
            self.__type = __type
            self.changed()

    @property
    def sx(self):
        """Source position on vertical axis(X)"""
        return self.__sx

    @sx.setter
    def sx(self, value):
        assert isinstance(value, float)
        if self.__sx != value:
            self.__sx = value
            self.changed()

    @property
    def sy(self):
        """Source position on vertical axis(Y)"""
        return self.__sy

    @sy.setter
    def sy(self, value):
        assert isinstance(value, float)
        if value != self.__sy:
            self.__sy = value
            self.changed()

    @property
    def dist(self):
        """Source distance in meters"""
        return self.__dist

    @dist.setter
    def dist(self, dist):
        assert isinstance(dist, float)
        if self.__dist != dist:
            self.__dist = dist
            self.changed()

    def to_dict(self):
        _dict = {
            "TYPE": self.type.name[0],
            "SX": self.sx,
            "SY": self.sy,
            "DIST": self.dist,
        }
        _dict.update(self.unmanaged_params)
        return _dict

    @staticmethod
    def from_dict(_dict):
        params = BeamGeoRP()
        params.load_from_dict(_dict)
        return params

    def load_from_dict(self, _dict):
        self._load_unmanaged_params(_dict)
        self.type = self._get_beam_geo_frm_letter(_dict["TYPE"])
        self.sx = _dict["SX"]
        self.sy = _dict["SY"]
        self.dist = _dict["DIST"]

    @staticmethod
    def _get_beam_geo_frm_letter(letter):
        if letter.lower() == "p":
            return BeamGeoType.parallel
        elif letter.lower() == "f":
            return BeamGeoType.fan
        elif letter.lower() == "c":
            return BeamGeoType.conical
        else:
            raise ValueError("Invalid beam geometry type (%s)" % letter)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["PJ.Gouttenoire", "H. Payno"]
__license__ = "MIT"
__date__ = "22/01/2018"

from tomwer.core.process.task import Task
import tomwer.version as tomwer_version
from tomwer.core.utils.pyhstutils import _getPyHSTDir
from tomwer.core.process.reconstruction.ftseries.params.paganin import PaganinMode
from tomwer.core.process.reconstruction.ftseries.params.ft import FixedSliceMode
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.settings import get_lbsram_path, get_dest_path
from silx.io.url import DataUrl
import logging
from tomwer.core import settings
from tomoscan.unitsystem import metricsystem
from .pyhstnxconverter import convert_nx_to_pyhst_hdf5, _pyhst_hdf5_info
import socket
import subprocess
import typing
import os

logger = logging.getLogger(__name__)


class PyHSTCaller(Task, input_names=("data",), output_names=("data",)):
    """Simple class containing information to be able to call PyHST"""

    REC_EXT = ".rec"
    """Extension of the record file"""
    PAR_EXT = ".par"
    """Extension of the reconstruction parameters file"""

    def __init__(self, varinfo=None, inputs=None, label=None):
        super().__init__(varinfo=varinfo, inputs=inputs, label=label)
        self._scan = inputs.get("data", None)
        self._dry_run = inputs.get("dry_run", False)
        self._rec_slices_only = inputs.get("rec_slices_only", True)
        self._full_rec_to_dest = inputs.get("full_rec_to_dest", True)
        self._slices_rec_to_dest = inputs.get("slices_rec_to_dest", False)
        if "recons_params" in inputs:
            raise KeyError(
                "Do not use recons_params but `pyhst2_recons_params` instead"
            )
        recons_params = inputs.get("pyhst2_params", None)
        if isinstance(recons_params, dict):
            self._recons_params = ReconsParams.from_dict(recons_params)
        else:
            self._recons_params = recons_params
        if self._recons_params is None:
            raise ValueError(
                "'pyhst2_recons_params' (ftseries reconstruction parameters) should be provided"
            )
        if self._recons_params is None:
            raise ValueError("'data' (scan) should be provided")

        self.execDir = _getPyHSTDir()

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "PyHST2"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return "defined in configuration"

    @staticmethod
    def definition():
        """definition of the process"""
        return "call PyHST2 to run tomography reconstructions"

    @staticmethod
    def isvalid(exec_dir, pyhst_exe):
        """

        :param exec_dir: directory containing the pyhst executable(s)
        :param pyhst_exe: name of the selected pyhst executable
        :returns: True if the exe file exists
        :rtype: bool
        """
        if pyhst_exe in os.environ.keys():
            return os.path.isfile(os.environ[pyhst_exe])
        elif exec_dir is None or pyhst_exe is None:
            return False
        else:
            return os.path.isfile(os.path.join(exec_dir, pyhst_exe))

    def _get_basefile_name_with_pag_des(
        self, paganin_opts: dict, scan_path: str, slice: typing.Union[None, int]
    ) -> str:
        """Return the *.par, *.rec and the *_correct.txt file name including
        the description of the paganin reconstruction

        :param several_pag: is there several paganin reconstruction to run
        :type: bool
        :param paganin_opts: paganin options for this reconstruction
        :type: dict
        :param scan_path: scan path
        :type: str
        :param slice: slice to reconstruct. If None then we want to reconstruct
                      the full volume
        :type: Union[None, int]
        :return: basename to use for file reconstruction
        :rtype: str
        """
        assert "DB" in paganin_opts, "requested to determine basefile name"
        assert "DB2" in paganin_opts, "requested to determine basefile name"
        assert "MODE" in paganin_opts, "requested to determine basefile name"
        scan_basename = os.path.basename(scan_path)
        pag_mod = PaganinMode.from_value(paganin_opts["MODE"])

        if pag_mod == PaganinMode.off:
            raise ValueError(
                "This function should not be call if paganin mode" " is off"
            )
        elif pag_mod in (PaganinMode.on, PaganinMode.both):
            if slice is None:
                try:
                    return "_".join(
                        (scan_basename + "pag", "%04d" % int(paganin_opts["DB"]))
                    )

                except Exception:
                    return "_".join(
                        (scan_basename + "pag", "%04.4i" % float(paganin_opts["DB"]))
                    )
            else:
                try:
                    return "_".join(
                        (scan_basename + "slice_pag", "%04d" % int(paganin_opts["DB"]))
                    )
                except Exception:
                    return "_".join(
                        (scan_basename + "slice_pag%04.4i" % float(paganin_opts["DB"]))
                    )

        elif pag_mod == PaganinMode.multi:
            if slice is None:
                multi_pag_val = "db%04.4i-db%04.4i" % (
                    paganin_opts["DB"],
                    float(paganin_opts["DB2"]),
                )
                return "_".join((scan_basename + "multipag", multi_pag_val))
            else:
                multi_pag_val = "db%04.4i-db%04.4i" % (
                    paganin_opts["DB"],
                    paganin_opts["DB2"],
                )
                return "_".join((scan_basename + "slice_multipag", multi_pag_val))

    def run(self):
        """
        The goal of this section is to generate all requested .rec and .par
        files that will ce executed later.

        Create the .par file and execute pyhst on it.
        Tomwer is looking one some specific folders to find the pyhst
        installation directory (first 'PYHST_DIR' if this environment variable
        is defined then /usr/bin. Get also pysht version defined as an
        environment variable like 'PyHST2_*')

        :param scan: acquisition to reconstruct
        :type: :class:`TomoBase`
        :param dry_run: if True then skip call to pyhst and only create .par
                        and .rec files
        :type: bool
        :param rec_slices_only: if True only call pyhst for the slice to be
                                reconstructed if any.
        :type: bool
        :param full_rec_to_dest: replace LBSRAM_ID by DEST_ID on the .par files
                                 used for full volume reconstruction
        :type: bool
        :param slices_rec_to_dest: replace LBSRAM_ID by DEST_ID on the .par
                                   files used for full slices reconstruction
        :type: bool
        """
        if self._scan is None:
            return None
        elif not isinstance(self._scan, TomwerScanBase):
            raise TypeError(
                "scan is expected to be an instance of "
                "TomwerScanBase. Not {}".format(type(self._scan))
            )
        self._scan.ftseries_recons_params = self._recons_params
        cuda_devices = self._scan.ftseries_recons_params.pyhst.cuda_devices

        def filter_paganin_values(rp):
            """We can have some value store in db and db2 but without a
            paganin mode turn on. If this is the case we want to avoid
            calling several time the same .par file"""
            assert rp is not None
            assert rp.paganin.mode in PaganinMode
            if rp.paganin.mode is PaganinMode.multi:
                pass
            elif rp.paganin.mode in (PaganinMode.on, PaganinMode.both):
                rp.paganin.db2 = "(-1, )"
            elif rp.paganin.mode is PaganinMode.off:
                rp.paganin.db = "(-1, )"
                rp.paganin.db2 = "(-1, )"
            return rp

        # cuda devices to use
        recons_params = ReconsParams(empty=False)
        recons_params.copy(self._scan.ftseries_recons_params)
        recons_params = filter_paganin_values(recons_params)
        recons_list = recons_params.to_unique_recons_set(as_to_dict=True)
        # set of unique reconstruction parameters
        has_several_pag = recons_params.paganin.has_several_db_param()

        # if needed convert hdf5 files
        if isinstance(self._scan, HDF5TomoScan):
            hdf5_info = convert_nx_to_pyhst_hdf5(scan=self._scan)
        else:
            hdf5_info = None

        slices = set((None,))
        if recons_params.ft.do_test_slice:
            fixed_slice = self._scan.ftseries_recons_params.ft.fixed_slice_as_list()
            try:
                fixed_slice = FixedSliceMode.from_value(fixed_slice)
            except (ValueError, TypeError):
                # in this case we are in row n and the value set is the
                # value(s) of the slice to reconstruct
                pass

            if fixed_slice is FixedSliceMode.middle:
                n_slice = self._scan.dim_2 or 2048
                slices.add(n_slice // 2)
            else:
                # in this case we are on row n mode
                if isinstance(fixed_slice, typing.Iterable):
                    [slices.add(slice) for slice in fixed_slice]
                else:
                    slices.add(fixed_slice)
        recons_urls = []
        for slice in slices:
            for recons_param in recons_list:
                if slice is None and self._rec_slices_only is True:
                    run_recons = False
                else:
                    run_recons = True
                if slice is None:
                    convert_lbsram_to_dest = self._full_rec_to_dest
                else:
                    convert_lbsram_to_dest = self._slices_rec_to_dest
                recons_urls.extend(
                    self._process_frm_one_slice_one_rp_set(
                        scan=self._scan,
                        slice=slice,
                        recons_param=recons_param,
                        has_several_pag=has_several_pag,
                        dry_run=self._dry_run,
                        run_recons=run_recons,
                        cuda_devices=cuda_devices,
                        convert_lbsram_to_dest=convert_lbsram_to_dest,
                        hdf5_info=hdf5_info,
                    )
                )
        self._scan.set_latest_reconstructions(recons_urls)

        if self._dry_run is False:
            entry = "entry"
            if isinstance(self._scan, HDF5TomoScan):
                entry = self._scan.entry
            with self._scan.acquire_process_file_lock():
                self.register_process(
                    process_file=self._scan.process_file,
                    entry=entry,
                    configuration=recons_params.to_dict(),
                    results={},
                    process_index=self._scan.pop_process_index(),
                    overwrite=True,
                )
        return self._scan

    def _process_frm_one_slice_one_rp_set(
        self,
        scan: TomwerScanBase,
        slice: typing.Union[int, None],
        recons_param: dict,
        cuda_devices: typing.Union[None, typing.Iterable],
        has_several_pag: bool,
        run_recons: bool,
        convert_lbsram_to_dest: bool,
        dry_run: bool,
        hdf5_info: _pyhst_hdf5_info,
    ) -> list:
        """
        process pyhstcaller from one specific slice indice and one set
        of reconstruction parameter (so all paramter should be unique)

        :param TomwerScanBase scan: scan to reconstruct
        :param Union[int, None] slice: slice index or None. If None then we
                                       will process the entire volume.
        :param recons_param: reconstruction parameters. warning: this set of
                             parameter should only contain unique values
                             contrary to the one given in process. This peace
                             of code won't apply the `to_unique_recons_set`
                             function
        :type recons_param: dict
        :param Union[None, Iterable] cuda_devices: list of cuda devices to use
                                                   for reconstruction
        :param bool has_several_pag: should we run several Paganin
                                     reconstruction for this scan.
        :param bool run_recons: if True then call pyhst within the created .par
                                file (depending on dry-run parameter of course)
        :param bool convert_lbsram_to_dest: if True then convert all path to
                                            the 'dest folder'.
        :param bool dry_run: if True then skip call to pyhst and only create
                             .par and .rec files
        :param hdf5_info: information regarding hdf5 if any created
        :return: list of the url to access the slice reconstructed
        :rtype: list
        """
        assert isinstance(recons_param, dict)
        assert isinstance(slice, (type(None), int))
        if slice is None:
            logger.info("manage .par for volume")
        else:
            logger.info("manage .par for slice %s" % slice)
        # deduce file base name
        if PaganinMode.from_value(recons_param["PAGANIN"]["MODE"]) == PaganinMode.off:
            base_file_name = os.path.basename(scan.path)
            if slice is not None:
                base_file_name = base_file_name + "slice"
                par_file_name = base_file_name + ("_%04d.par" % slice)
            else:
                par_file_name = base_file_name

            if not par_file_name.endswith(".par"):
                par_file_name = par_file_name + ".par"
            correct_file_name = os.path.join(scan.path, "correct.txt")
        else:
            base_file_name = self._get_basefile_name_with_pag_des(
                paganin_opts=recons_param["PAGANIN"], scan_path=scan.path, slice=slice
            )
            # note: adding the slice on the name file is done on the
            # `_get_basefile_name_with_pag_des` function
            correct_file_name = base_file_name + "_correct.txt"
            if slice is None:
                par_file_name = base_file_name + ".par"
            else:
                par_file_name = base_file_name + "_%04d.par" % slice

        # create correct.txt, .par, .rec and execute the .par using physt
        try:
            if recons_param["FTAXIS"]["DO_AXIS_CORRECTION"] and (
                recons_param["FTAXIS"]["USE_TOMWER_AXIS"]
                or recons_param["FTAXIS"]["TRY_USE_OLD_TOMWER_AXIS"]
            ):
                cor = self.get_center_of_rotation(
                    scan=scan,
                    use_tomwer_axis=recons_param["FTAXIS"]["USE_TOMWER_AXIS"],
                    use_old_tomwer_processes=recons_param["FTAXIS"][
                        "TRY_USE_OLD_TOMWER_AXIS"
                    ],
                    output_file=correct_file_name,
                )
                cor = cor if cor is None else float(cor)
            else:
                cor = None
            self.make_par_file(
                dirname=scan.path,
                scan=scan,
                recons_param=recons_param,
                out_base_file_name=base_file_name,
                slice=slice,
                convert_path=convert_lbsram_to_dest,
                par_file_path=os.path.join(scan.path, par_file_name),
                cor=cor,
                hdf5_info=hdf5_info,
            )
            assert os.path.exists(os.path.join(scan.path, par_file_name))
        except ValueError as e:
            logger.error("Fail to generate .par file. " + str(e))
            return list()
        else:
            if recons_param["PYHSTEXE"]["MAKE_OAR_FILE"]:
                # create .rec if needed
                prefix = base_file_name or os.path.basename(scan.path)
                rec_file_name = os.path.join(scan.path, prefix + PyHSTCaller.REC_EXT)
                self.make_rec_file(
                    rec_file=rec_file_name,
                    recons_params=recons_param,
                    par_file=par_file_name,
                )

            # execute pyhst
            exec_dir = _getPyHSTDir()
            pyhst_exe = recons_param["PYHSTEXE"]["EXE"]
            if recons_param["PYHSTEXE"]["VERBOSE"]:
                pyhst_out = None
            else:
                pyhst_out = os.path.join(
                    scan.path, recons_param["PYHSTEXE"]["VERBOSE_FILE"]
                )
            if run_recons is True:
                if isinstance(scan, HDF5TomoScan):
                    exec_path = scan.path
                else:
                    exec_path = None
                try:
                    self.execute(
                        exec_dir=exec_dir,
                        pyhst_exe=pyhst_exe,
                        par_file=os.path.join(scan.path, par_file_name),
                        cuda_devices=cuda_devices,
                        dry_run=dry_run,
                        pyhst_out=pyhst_out,
                        exec_path=exec_path,
                    )
                except ValueError as e:
                    logger.error(e)
                    return list()
                else:
                    if dry_run is True:
                        return list()
                    else:
                        if recons_param["FT"]["VOLOUTFILE"] == 1:
                            output_file_ext = ".vol"
                            scheme = "tomwer"
                        else:
                            output_file_ext = ".edf"
                            scheme = "fabio"
                        recons_file = par_file_name.replace(".par", output_file_ext)
                        recons_file = os.path.join(scan.path, recons_file)
                        recons_url = DataUrl(
                            file_path=recons_file, data_slice=None, scheme=scheme
                        )
                        return [recons_url]
            else:
                return list()

    def get_center_of_rotation(
        self,
        scan: TomwerScanBase,
        use_tomwer_axis: bool,
        use_old_tomwer_processes: str,
        output_file: str,
    ):
        """

        :param scan: scan for which we want to create the correct file.
                     if `USE_TOMWER_AXIS` is True then will first look for some
                     axis process computation in the history.
                     if `TRY_USE_OLD_TOMWER_AXIS` is True then will look for
                     any tomwer_processes file store in the scan directory
                     to extract it the center of rotation value
        :type: TomoBase
        :param use_tomwer_axis: use the history contained in the axis file
        :type: bool
        :param use_old_tomwer_processes: use information contained in the
                                         tomwer_processes file
        :type: bool
        :param output_file: output file name
        :type: str
        :return: name of the file where the center of rotation information
                 has been written. None if not cor information stored
        :type: Union[None,str]
        """
        if use_tomwer_axis:
            if (
                scan.axis_params is not None
                and scan.axis_params.relative_cor_value is not None
            ):
                logger.info(
                    "load center of rotation from previous axis "
                    "process: %s" % scan.axis_params.relative_cor_value
                )
                return scan.axis_params.relative_cor_value
        if use_old_tomwer_processes and scan.process_file is not None:
            if os.path.exists(scan.process_file):
                from tomwer.core.process.reconstruction.axis import AxisRP

                cor = AxisRP.get_cor_position_frm_tomwer_file(
                    file_path=scan.process_file
                )
                logger.info(
                    "load center of rotation from %s: %s" % (scan.process_file, cor)
                )
                return cor
        return None

    def execute(
        self,
        exec_dir: str,
        pyhst_exe: str,
        par_file: str,
        cuda_devices: list,
        dry_run: bool = False,
        pyhst_out: typing.Union[None, str] = None,
        exec_path=None,
    ) -> bool:
        """
        Run pyhst for the given par file

        :param exec_dir: directory containing the pyhst executable(s)
        :param str pyhst_exe: name of the selected pyhst executable
        :param str par_file: file to use to call pyhst
        :param list cuda_devices: list of `CudaDevice` to use in localhost for
                             executing the reconstruction
        :param bool dry_run: if True, won't execute the pyhst reconstruction
        :param pyhst_out: if specify, link to pyhst log file
        :rtype: Union[None,str]
        """
        assert par_file is not None
        if dry_run is False:
            if not os.path.exists(par_file):
                raise IOError(par_file + "does not exists")

            if not PyHSTCaller.isvalid(pyhst_exe=pyhst_exe, exec_dir=exec_dir):
                raise ValueError(
                    "physt executable definition, %s, %s is invalid"
                    % (exec_dir, pyhst_exe)
                )

        try:
            if pyhst_exe in os.environ.keys():
                pyhst_exe = os.environ[pyhst_exe]
            gpu_cmd_opt = self._get_gpu_command(cuda_devices)
            logger.info(" ".join(("launching: ", pyhst_exe, par_file, gpu_cmd_opt)))
            if dry_run is False:
                if pyhst_out is not None:
                    with open(pyhst_out, "w") as fout:
                        self.__execute(
                            pyhst_exe=pyhst_exe,
                            par_file=par_file,
                            gpu_cmd_opt=gpu_cmd_opt,
                            stdout=fout,
                            exec_path=exec_path,
                        )
                else:
                    self.__execute(
                        pyhst_exe=pyhst_exe,
                        par_file=par_file,
                        gpu_cmd_opt=gpu_cmd_opt,
                        stdout=None,
                        exec_path=exec_path,
                    )
        except OSError as e:
            logger.error("reconstruction fails", e)
            return False
        else:
            logger.info("reconstruction of %s finished." % par_file)
            return True

    def __execute(self, pyhst_exe, par_file, gpu_cmd_opt, stdout, exec_path):
        logger.info("processing pyhst reconstruction... (%s)" % par_file)
        subprocess.call(
            " ".join((pyhst_exe, par_file, gpu_cmd_opt)),
            shell=True,
            stdout=stdout,
            stderr=stdout,
            cwd=exec_path,
        )
        logger.info("pyhst reconstruction finished (%s)" % par_file)

    @staticmethod
    def _get_gpu_command(devices):
        """compute the extension of the command to use specific gpus"""
        if devices is None:
            return ""
        else:
            devices_id = []
            for device in devices:
                from tomwer.core.utils.device import CudaDevice

                if isinstance(device, CudaDevice):
                    devices_id.append(str(device.id))
                elif isinstance(device, int):
                    devices_id.append(str(device))
            devices_id.insert(0, socket.gethostname())
            return ",".join(devices_id)

    def make_par_file(
        self,
        dirname,
        scan,
        recons_param,
        slice,
        convert_path,
        par_file_path: str,
        hdf5_info,
        out_base_file_name=None,
        options=None,
        cor=None,
    ):
        """
        Function creating the .par file

        creates parameter file for processing with PyHST
        The par file requires 9 fixed arguments:

        * argument 1: direc : path of the scan directory
        * argument 2: prefix : scan prefix
        * argument 3: nvue : number of angles used in the reconstruction
        * argument 4: num_image_1 : number of pixels in the projections along the first axis (horizontal)
        * argument 5: num_image_2 : number of pixels in the projections along the second axis (vertical)
        * argument 6: image_pixel_size : pixel size in microns
        * argument 7: corr : do flatfield correction + log or not: 'YES' or 'NO'
        * argument 8: refon : interval between the flatfield files: e.g. 100
        * argument 9: offset : position of rotation axis with respect to middle in pixels

        :param dirname: the directory to store the .par file
        :param TomwerScanBase scan: scan to reconstruct
        :param slice: slice to reconstruct. If None then we will reconstruct
                      the entire volume
        :type: int
        :param convert_path: if true, convert all path to destination path
                             (used on lbsram system)
        :type: bool
        :param out_base_file_name: the par file name.
        :type: Union[None,str]
        :param correct_file: text file containing the correction to apply
        :type: str
        :param options: dictionary with options you want to set.
                        Regular options are:

                        * 'parfilename': parameter file ( default: direc/prefix.par )
                        * 'start_voxel_1': first pixel in x (starting from 1)
                          (default 1)
                        * 'start_voxel_2': first pixel in y (starting from 1)
                          (default 1)
                        * 'start_voxel_3': first pixel in z (starting from 1)
                          (default 1)
                        * 'end_voxel_1': last pixel in x (starting from 1)
                          (default num_image_1)
                        * 'end_voxel_2': last pixel in y (starting from 1)
                          (default num_image_1)
                        * 'end_voxel_3': last pixel in z (starting from 1)
                          (default num_image_2)
                        * 'angle_offset': offset angle in degrees
                          (default 0)
                        * 'output_file': name of the volume file
                          (default: direc/prefix.vol)
                        * 'ff_prefix': prefix of the flatfield images
                          (default: direc/refHST)
                        * 'background_file': name of the dark file
                          (default: = direc/dark.edf)
                        * 'horaxis': rotation axis horizontal (=constant y) (1)
                          or not (=constant x) (0) ( default 0 )
                        * 'correct_spikes_threshold' : threshold value for ccd
                          filter when value is given, ccd filter will be
                          applied with this parameter (default : parameter is
                          set to 0.04, but DO_CCD_FILTER = 'NO')
                        * 'ccd_filter': filter for correction of spikes
                          (default: 'CCD_Filter')
                        * 'ccd_filter_para': parameters for ccd filter when
                          given overrides choice with correct_spikes_threshold
                        * 'do_axis_correction' : correct for sample motion
                          (default: 'NO')
                        * 'axis_correction_file': name of file with sample
                          movement (default: 'correct.txt' )
                        * 'first column': x motion [ , second column: y motion]
                        * 'do_sino_filter': correction of rings by filtering
                          the sinogram ( default: 'NO' )
                        * 'sino_filter': filter for correction of rings
                          (default: 'SINO_Filter')
                        * 'correct_rings_nb': number of elements (`|f|>0`) that
                          will not be filtered from sinograms (default 8)
                        * 'padding': padding at edges with edge values 'E' or
                          zero '0' ( default: 'E' )
                        * 'axis_to_the_center' : move axis to the center 'Y' or
                          not 'N' (default: 'Y') !!!
                        * 'angle_between_projections': angle between successive
                          projections ( default 180/nvue )
                        * 'doubleffcorrection': name of the file (e.g.
                          filtered_mean.edf) for double flatfield subtracted
                          after normal flatfield and logarithm.
                        * 'do_projection_median': calculate median of
                          projections (after flat/log) (default: 'NO')
                        * 'projection_median_filename' : name of output
                          edf-file with median (default: median.edf)
                        * 'do_projection_mean': calculate mean of projections
                          (after flat/log) (default: 'NO')
                        * 'projection_mean_filename': name of output edf-file
                          with mean (default: mean.edf) PyHST uses for both
                          median and mean PROJECTION_MEDIAN_FILENAME

                    Advanced options (= do not change, except if you know what
                    you are doing):

                        * 'zeroclipvalue': minimum value of radiographs after
                          flatfield / before logarithm values below are clipped
                          (default 1e-9)
                        * 'oneclipvalue': maximum value of radiographs after
                          flatfield / before logarithm values above are clipped
                          (default: not applied)
                        * 'multiframe': more than one projection in input file
                          (default: 0)
                        * 'num_first_image': index of first image used in
                          reconstruction (default 0)
                        * 'num_last_image': index of last image
                          (default nvue-1)
                        * 'number_length_varies': width index varies or not
                          (default: 'NO')
                        * 'length_of_numerical_part': width of image index
                          (default 4)
                        * 'file_postfix': extension of the image files
                          (default: '.edf')
                        * 'file_interval': downsample in angle (default 1)
                        * 'subtract_background': subtract dark file 'YES' or
                          not 'NO' ( default: corr )
                        * 'correct_flatfield': divide by flatfield image 'YES'
                          or not 'NO' (default: corr)
                        * 'take_logarithm': use -log 'Y' or 'N'
                          (default: 'YES')
                        * 'output_sinograms': output sinogram files or not
                          (default: 'NO')
                        * 'output_reconstruction': do the reconstruction or not
                          (default: 'YES')
                        * 'oversampling_factor': oversampling of projections
                          before backprojection (0 = linear interpolation,
                          1 = nearest pixel) ( default 4 )
                        * 'sinogram_megabytes': obsolete maximum memory in MB used for
                          reading the sinograms ( default 400 )
                        * 'cache_kilobytes': obsolete ( default 256 )
                        * 'display_graphics': obsolete NOT IMPLEMENTED IN PYHST
                          (default: 'NO')
                        * cor: center of rotation, given from [-image_width/2, image_width/2]
            :param hdf5_info: information regarding hdf5 if any created
        """
        assert isinstance(scan, TomwerScanBase)
        # add some missing values:
        recons_param["FT"]["FILE_PREFIX"] = os.path.basename(scan.path)
        # TODO: how can we obtain num of last image, NUM_IMAGE_1, NUM_IMAGE_2... ???
        recons_param["FT"]["NUM_LAST_IMAGE"] = len(scan.projections)
        recons_param["FT"]["NUM_IMAGE_1"] = scan.dim_1
        recons_param["FT"]["NUM_IMAGE_2"] = scan.dim_2
        scan_range = scan.scan_range
        recons_param["FT"]["IMAGE_PIXEL_SIZE_1"] = (
            (scan.pixel_size / metricsystem.micrometer.value)
            if scan.pixel_size is not None
            else None
        )
        # TODO: the next values should be deduced from some 'flow' recording or scan tag
        recons_param["FT"]["CORRECT_FLATFIELD"] = True
        recons_param["FT"]["FF_FILE_INTERVAL"] = scan.ff_interval

        # those values can't be set by options
        prefix = recons_param["FT"]["FILE_PREFIX"]
        nvue = recons_param["FT"]["NUM_LAST_IMAGE"]
        num_image_1 = recons_param["FT"]["NUM_IMAGE_1"]
        num_image_2 = recons_param["FT"]["NUM_IMAGE_2"]
        image_pixel_size = recons_param["FT"]["IMAGE_PIXEL_SIZE_1"]
        corr = "YES" if recons_param["FT"]["CORRECT_FLATFIELD"] else "NO"
        refon = recons_param["FT"]["FF_FILE_INTERVAL"]
        direc = dirname

        fileParameter = self._loadDefaultPar(
            recons_param=recons_param,
            out_base_file_name=out_base_file_name,
            direc=direc,
            image_pixel_size=image_pixel_size,
            corr=corr,
            nvue=nvue,
            refon=refon,
            num_image_1=num_image_1,
            num_image_2=num_image_2,
            slice=slice,
            scan_range=scan_range,
            convert_lbsram=convert_path,
            cor=cor,
        )
        # Dealing with options
        if options:
            for opt in options:
                fileParameter[opt] = options[opt]

        not_available = "N.A."

        FT = recons_param["FT"]
        PAG = recons_param["PAGANIN"]
        BG = recons_param["BEAMGEO"]
        paganin = PaganinMode.from_value(recons_param["PAGANIN"]["MODE"])

        length_of_numerical_part = FT["NUM_PART"]

        ###################################################
        ##### check and calculate dependent variables #####
        ###################################################
        if fileParameter["horaxis"]:
            # move center of rotation to the pyhst referrence [0, image_width]
            if cor is None:
                cor = num_image_2 / 2 if num_image_2 else None
            else:
                cor = num_image_2 / 2 + cor
            fileParameter["rotation_vertical"] = "NO"
            fileParameter["rotation_axis_position"] = cor
            fileParameter["end_voxel_12_limit "] = num_image_2
            fileParameter["end_voxel_3_limit"] = num_image_1
        else:
            # move center of rotation to the pyhst referrence [0, image_width]
            if cor is None:
                cor = num_image_1 / 2 if num_image_1 else None
            else:
                cor = num_image_1 / 2 + cor
            fileParameter["rotation_vertical"] = "YES"
            fileParameter["rotation_axis_position"] = cor
            fileParameter["end_voxel_12_limit"] = num_image_1
            fileParameter["end_voxel_3_limit"] = num_image_2

        # start_voxel should be > 0 ; end_voxel should be <= image dimension
        if fileParameter["start_voxel_1"] < 1:
            logger.warning("START_VOXEL_1 can not be zero or negative, forcing to 1")
            fileParameter["start_voxel_1"] = 1
        if fileParameter["start_voxel_2"] < 1:
            logger.warning("START_VOXEL_2 can not be zero or negative, forcing to 1")
            fileParameter["start_voxel_2"] = 1
        if fileParameter["start_voxel_3"] < 1:
            logger.warning("START_VOXEL_3 can not be zero or negative, forcing to 1")
            fileParameter["start_voxel_3"] = 1

        if fileParameter["half_acquisition"] == 0:
            if (fileParameter["end_voxel_1"] is not None) and (
                fileParameter["end_voxel_1"] > fileParameter["end_voxel_12_limit"]
            ):
                logger.warning(
                    "END_VOXEL_1 can not be larger than image dimension, forcing to maximum"
                )
                fileParameter["end_voxel_1"] = fileParameter["end_voxel_12_limit"]

            if (fileParameter["end_voxel_2"] is not None) and (
                fileParameter["end_voxel_2"] > fileParameter["end_voxel_12_limit"]
            ):
                logger.warning(
                    "END_VOXEL_2 can not be larger than image dimension, forcing to maximum"
                )
                fileParameter["end_voxel_2"] = fileParameter["end_voxel_12_limit"]

            fileParameter["avoidhalftomo"] = "Y"
        else:
            fileParameter["avoidhalftomo"] = "N"

        if "FORCE_HALF_ACQ" in recons_param["FT"].keys():
            if FT["FORCE_HALF_ACQ"] == 1 and fileParameter["half_acquisition"] != 0:
                fileParameter["avoidhalftomo"] = "-1"

        if (
            (fileParameter["end_voxel_3"] is not None)
            and (fileParameter["end_voxel_3_limit"] is not None)
            and (fileParameter["end_voxel_3"] > fileParameter["end_voxel_3_limit"])
        ):
            logger.warning(
                "END_VOXEL_3 can not be larger than image dimension, forcing to maximum\n"
            )
            fileParameter["end_voxel_3"] = fileParameter["end_voxel_3_limit"]

        # Remove directory head if necessary
        if convert_path:
            fileParameter["output_file"] = self._remove_lbs_head(
                fileParameter["output_file"]
            )

        if slice is None:
            if recons_param["FT"]["VOLOUTFILE"] == 1:
                output_file_ext = ".vol"
            else:
                output_file_ext = ".edf"
        else:
            output_file_ext = ""

        fileParameter["output_file"] = fileParameter["output_file"] + output_file_ext

        if paganin == PaganinMode.multi and fileParameter["multipag_extra_files"] == 1:

            # Remove directory head if necessary
            fileParameter["extra_output_file"] = self._remove_lbs_head(
                out_base_file_name
            )

            if fileParameter["multipag_keep_bone"] == 1:
                fileParameter["output_file_bone"] = (
                    fileParameter["extra_output_file"] + "_bone.vol"
                )

            if fileParameter["multipag_keep_soft"] == 1:
                fileParameter["output_file_soft"] = (
                    fileParameter["extra_output_file"] + "_soft.vol"
                )

            if fileParameter["multipag_keep_mask"] == 1:
                fileParameter["output_file_mask"] = (
                    fileParameter["extra_output_file"] + "_mask.vol"
                )

            if fileParameter["multipag_keep_abs"] == 1:
                fileParameter["output_file_abs"] = (
                    fileParameter["extra_output_file"] + "_abs.vol"
                )

            if fileParameter["multipag_keep_corr"] == 1:
                fileParameter["output_file_corr"] = (
                    fileParameter["extra_output_file"] + "_corr.vol"
                )

        if fileParameter["parfilename"][0] != os.path.sep:
            fileParameter["parfilename"] = os.path.join(
                direc, fileParameter["parfilename"]
            )

        if fileParameter["background_file"][0] != os.path.sep:
            fileParameter["background_file"] = os.path.join(
                direc, fileParameter["background_file"]
            )

        if fileParameter["ff_prefix"][0] != os.path.sep:
            fileParameter["ff_prefix"] = os.path.join(direc, fileParameter["ff_prefix"])

        #############################
        #####    angle offset   #####
        #############################
        if not ("ANGLE_OFFSET" in FT):
            fileParameter["angle_offset"] = 0.0

        #############################
        ##### correction spikes #####
        #############################
        if not ("CORRECT_SPIKES_THRESHOLD" in FT):
            fileParameter["correct_spikes_threshold"] = 0.0
        if not ("CCD_FILTER_PARA" in FT):
            fileParameter["ccd_filter_para"] = {}
        if not ("DO_CCD_FILTER" in FT):
            fileParameter["do_ccd_filter"] = "NO"

        if fileParameter["ccd_filter_para"] == "CCD_Filter":
            if "correct_spikes_threshold" not in fileParameter:
                fileParameter["do_ccd_filter_preference"] = "NO"
                fileParameter["ccd_filter_para"] = {
                    "threshold": fileParameter["correct_spikes_threshold_default"]
                }
            elif fileParameter["correct_spikes_threshold"].isinf():
                fileParameter["do_ccd_filter"] = "NO"
                fileParameter["ccd_filter_para"] = {
                    "threshold": fileParameter["correct_spikes_threshold_default"]
                }
            else:
                fileParameter["do_ccd_filter_preference"] = "YES"
                fileParameter["ccd_filter_para"] = {
                    "threshold": fileParameter["correct_spikes_threshold"]
                }
        else:
            fileParameter["do_ccd_filter_preference"] = "YES"

        if len(fileParameter["do_ccd_filter"]) == 0:
            fileParameter["do_ccd_filter"] = fileParameter["do_ccd_filter_preference"]

        # TODO for now we call several time the reconstructor but no dircet multiframe
        fileParameter["multiframe"] = 0

        ##########################
        ##### create parfile #####
        ##########################

        with open(par_file_path, "wb") as fidparfile:
            file_prefix = os.path.join(direc, prefix)
            if convert_path:
                file_prefix = PyHSTCaller._remove_lbs_head(file_prefix)

            if fileParameter["multiframe"]:
                file_prefix = file_prefix + ".edf"
            fileParameter["file_prefix"] = file_prefix

            # Remove directory head if necessary
            if convert_path:
                fileParameter["background_file"] = self._remove_lbs_head(
                    dirin=fileParameter["background_file"]
                )
                fileParameter["ff_prefix"] = self._remove_lbs_head(
                    dirin=fileParameter["ff_prefix"]
                )

            if "PAG_LENGTH" in recons_param["PAGANIN"]:
                fileParameter["pag_length"] = recons_param["PAGANIN"]["DB"]
            fidparfile.write(bytes("# pyHST_SLAVE PARAMETER FILE\n\n", "UTF-8"))
            fidparfile.write(
                bytes(
                    "# generated by tomwer - %s\n\n" % tomwer_version.version, "UTF-8"
                )
            )

            fidparfile.write(
                bytes("# Parameters defining the projection file series\n", "UTF-8")
            )
            fidparfile.write(
                bytes("MULTIFRAME = %d\n\n" % fileParameter["multiframe"], "UTF-8")
            )
            if hdf5_info is None:
                _file_prefix = fileParameter["file_prefix"]
            else:
                _file_prefix = os.path.relpath(hdf5_info.projections.file, scan.path)
            fidparfile.write(bytes("FILE_PREFIX = %s\n" % _file_prefix, "UTF-8"))
            if hdf5_info is not None:
                fidparfile.write(
                    bytes(
                        "PROJ_DS_NAME = %s\n" % hdf5_info.projections.dataset_path,
                        "UTF-8",
                    )
                )

            fidparfile.write(
                bytes(
                    "NUM_FIRST_IMAGE = %d # No. of first projection file\n"
                    % fileParameter["num_first_image"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes(
                    "NUM_LAST_IMAGE = %d # No. of last projection file\n"
                    % fileParameter["num_last_image"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes(
                    "NUMBER_LENGTH_VARIES = %s\n"
                    % fileParameter["number_length_varies"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes(
                    "LENGTH_OF_NUMERICAL_PART = %d # No. of characters\n"
                    % length_of_numerical_part,
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes("FILE_POSTFIX = %s\n" % fileParameter["file_postfix"], "UTF-8")
            )
            fidparfile.write(
                bytes(
                    "FILE_INTERVAL = %d # Interval between input files\n"
                    % fileParameter["file_interval"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes("\n# Parameters defining the projection file format\n", "UTF-8")
            )
            if num_image_1 is not None:
                fidparfile.write(
                    bytes(
                        "NUM_IMAGE_1 = %d # Number of pixels horizontally\n"
                        % num_image_1,
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "NUM_IMAGE_1 = ??? # Number of pixels horizontally\n", "UTF-8"
                    )
                )
            if num_image_2 is not None:
                fidparfile.write(
                    bytes(
                        "NUM_IMAGE_2 = %d # Number of pixels vertically\n"
                        % num_image_2,
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes("NUM_IMAGE_2 = ??? # Number of pixels vertically\n", "UTF-8")
                )
            if image_pixel_size is not None:
                fidparfile.write(
                    bytes(
                        "IMAGE_PIXEL_SIZE_1 = %f # Pixel size horizontally (microns)\n"
                        % image_pixel_size,
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "IMAGE_PIXEL_SIZE_1 = ??? # Pixel size horizontally (microns)\n",
                        "UTF-8",
                    )
                )
            if fileParameter["image_pixel_size_2"] is not None:
                fidparfile.write(
                    bytes(
                        "IMAGE_PIXEL_SIZE_2 = %f # Pixel size vertically\n"
                        % fileParameter["image_pixel_size_2"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes("IMAGE_PIXEL_SIZE_2 = ??? # Pixel size vertically\n", "UTF-8")
                )
            fidparfile.write(
                bytes("\n# Parameters defining background treatment\n", "UTF-8")
            )

            if fileParameter["correct"]:
                fidparfile.write(
                    bytes(
                        "SUBTRACT_BACKGROUND = %s # Subtract background from data\n"
                        % fileParameter["subtract_background"],
                        "UTF-8",
                    )
                )
                if hdf5_info is None:
                    fidparfile.write(
                        bytes(
                            "BACKGROUND_FILE = %s\n" % fileParameter["background_file"],
                            "UTF-8",
                        )
                    )
                elif hdf5_info.dark is not None:
                    fidparfile.write(
                        bytes(
                            "BACKGROUND_FILE = %s\n"
                            % os.path.relpath(hdf5_info.dark.file, scan.path),
                            "UTF-8",
                        )
                    )
                if hdf5_info is not None and hdf5_info.dark is not None:
                    fidparfile.write(
                        bytes(
                            "BACKGROUND_DS_NAME = %s\n" % hdf5_info.dark.dataset_path,
                            "UTF-8",
                        )
                    )
            else:
                fidparfile.write(
                    bytes(
                        "SUBTRACT_BACKGROUND = %s # No background subtraction\n"
                        % fileParameter["subtract_background"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes("BACKGROUND_FILE = %s\n" % not_available, "UTF-8")
                )

            fidparfile.write(
                bytes("\n# Parameters defining flat-field treatment\n", "UTF-8")
            )
            if fileParameter["correct"]:
                fidparfile.write(
                    bytes(
                        "CORRECT_FLATFIELD = %s # Divide by flat-field image\n"
                        % fileParameter["correct_flatfield"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        "FLATFIELD_CHANGING = %s # Series of flat-field files\n"
                        % fileParameter["flatfield_changing"],
                        "UTF-8",
                    )
                )
                if hdf5_info is None:
                    fidparfile.write(
                        bytes(
                            "FLATFIELD_FILE = %s\n"
                            % os.path.basename(fileParameter["flatfield_file"]),
                            "UTF-8",
                        )
                    )
                    fidparfile.write(
                        bytes("FF_PREFIX = %s\n" % fileParameter["ff_prefix"], "UTF-8")
                    )
                elif hdf5_info.flats is not None:
                    _ff_files = ", ".join(
                        [
                            '"{}"'.format(os.path.relpath(file_name, scan.path))
                            for file_name in hdf5_info.flats.files
                        ]
                    )
                    _ff_files = "[{}]".format(_ff_files)
                    fidparfile.write(bytes("FF_PREFIX = %s\n" % _ff_files, "UTF-8"))
                    fidparfile.write(
                        bytes(
                            "FF_DS_NAME = %s\n" % hdf5_info.flats.dataset_path, "UTF-8"
                        )
                    )

                fidparfile.write(
                    bytes(
                        "FF_NUM_FIRST_IMAGE = %d # No. of first flat-field file\n"
                        % fileParameter["ff_num_first_image"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        "FF_NUM_LAST_IMAGE = %d # No. of last flat-field file\n"
                        % fileParameter["ff_num_last_image"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        "FF_NUMBER_LENGTH_VARIES = %s\n"
                        % fileParameter["number_length_varies"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        "FF_LENGTH_OF_NUMERICAL_PART = %d # No. of characters\n"
                        % length_of_numerical_part,
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes("FF_POSTFIX = %s\n" % fileParameter["file_postfix"], "UTF-8")
                )
                if fileParameter["ff_file_interval"] is not None:
                    if hdf5_info is None:
                        fidparfile.write(
                            bytes(
                                "FF_FILE_INTERVAL = %d # Interval between flat-field files\n"
                                % fileParameter["ff_file_interval"],
                                "UTF-8",
                            )
                        )
                    elif hdf5_info.flats is not None:
                        # note: in the case of hdf5 we write at most 2 flat files
                        fidparfile.write(
                            bytes(
                                "FF_FILE_INTERVAL = %d # Interval between flat-field files\n"
                                % len(scan.projections),
                                "UTF-8",
                            )
                        )

                else:
                    fidparfile.write(
                        bytes(
                            "FF_FILE_INTERVAL = ??? # Interval between flat-field files\n",
                            "UTF-8",
                        )
                    )
            else:
                fidparfile.write(
                    bytes(
                        "CORRECT_FLATFIELD = %s # No flat-field correction\n"
                        % fileParameter["correct_flatfield"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes("FLATFIELD_CHANGING = %s\n" % not_available, "UTF-8")
                )
                fidparfile.write(
                    bytes("FLATFIELD_FILE = %s\n" % not_available, "UTF-8")
                )
                fidparfile.write(bytes("FF_PREFIX = %s\n" % not_available, "UTF-8"))
                fidparfile.write(
                    bytes("FF_NUM_FIRST_IMAGE = %s\n" % not_available, "UTF-8")
                )
                fidparfile.write(
                    bytes("FF_NUM_LAST_IMAGE = %s\n" % not_available, "UTF-8")
                )
                fidparfile.write(
                    bytes("FF_NUMBER_LENGTH_VARIES = %s\n" % not_available, "UTF-8")
                )
                fidparfile.write(
                    bytes("FF_LENGTH_OF_NUMERICAL_PART = %s\n" % not_available, "UTF-8")
                )
                fidparfile.write(bytes("FF_POSTFIX = %s\n" % not_available, "UTF-8"))
                fidparfile.write(
                    bytes("FF_FILE_INTERVAL = %s\n" % not_available, "UTF-8")
                )
            if recons_param["FT"]["DFFCORRECTION"]:
                fidparfile.write(
                    bytes(
                        "DOUBLEFFCORRECTION = 1 # Apply double flat field correction\n",
                        "UTF-8",
                    )
                )
                sigma_2 = recons_param["FT"]["DFF_SIGMA"]
                fidparfile.write(
                    bytes(
                        "FF2_SIGMA = %s # double flat field high-pass filter\n"
                        % sigma_2,
                        "UTF-8",
                    )
                )

            fidparfile.write(
                bytes(
                    "\nTAKE_LOGARITHM = %s # Take log of projection values\n"
                    % fileParameter["take_logarithm"],
                    "UTF-8",
                )
            )
            if not (fileParameter["doubleffcorrection"] == []):
                fidparfile.write(
                    bytes(
                        'DOUBLEFFCORRECTION = "filtered_mean.edf" # double flatfield\n',
                        "UTF-8",
                    )
                )

            fidparfile.write(bytes("\n# Parameters defining experiment\n", "UTF-8"))
            fidparfile.write(
                bytes(
                    "ANGLE_BETWEEN_PROJECTIONS = %f # Increment angle in degrees\n"
                    % fileParameter["angle_between_projections"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes(
                    "ROTATION_VERTICAL = %s\n" % fileParameter["rotation_vertical"],
                    "UTF-8",
                )
            )
            if fileParameter["rotation_axis_position"] is not None:
                fidparfile.write(
                    bytes(
                        "ROTATION_AXIS_POSITION = %f # Position in pixels\n"
                        % fileParameter["rotation_axis_position"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "ROTATION_AXIS_POSITION = ??? # Position in pixels\n", "UTF-8"
                    )
                )
            fidparfile.write(bytes("\n# Parameters defining reconstruction\n", "UTF-8"))
            fidparfile.write(
                bytes(
                    "OUTPUT_SINOGRAMS = %s # Output sinograms to files or not\n"
                    % fileParameter["output_sinograms"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes(
                    "OUTPUT_RECONSTRUCTION = %s # Reconstruct and save or not\n"
                    % fileParameter["output_reconstruction"],
                    "UTF-8",
                )
            )
            if fileParameter["start_voxel_1"] is not None:
                fidparfile.write(
                    bytes(
                        "START_VOXEL_1 =   %4d # X-start of reconstruction volume\n"
                        % fileParameter["start_voxel_1"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "START_VOXEL_1 =   ??? # X-start of reconstruction volume\n",
                        "UTF-8",
                    )
                )
            if fileParameter["start_voxel_2"] is not None:
                fidparfile.write(
                    bytes(
                        "START_VOXEL_2 =   %4d # Y-start of reconstruction volume\n"
                        % fileParameter["start_voxel_2"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "START_VOXEL_2 =   ??? # Y-start of reconstruction volume\n",
                        "UTF-8",
                    )
                )
            if fileParameter["start_voxel_3"] is not None:
                fidparfile.write(
                    bytes(
                        "START_VOXEL_3 =   %4d # Z-start of reconstruction volume\n"
                        % fileParameter["start_voxel_3"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "START_VOXEL_3 =   ??? # Z-start of reconstruction volume\n",
                        "UTF-8",
                    )
                )
            if fileParameter["end_voxel_1"] is not None:
                fidparfile.write(
                    bytes(
                        "END_VOXEL_1 =   %4d # X- of reconstruction volume\n"
                        % fileParameter["end_voxel_1"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "END_VOXEL_1 =   %4d # X- of reconstruction volume\n", "UTF-8"
                    )
                )
            if fileParameter["end_voxel_2"] is not None:
                fidparfile.write(
                    bytes(
                        "END_VOXEL_2 =   %4d # Y- of reconstruction volume\n"
                        % fileParameter["end_voxel_2"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "END_VOXEL_2 =   ??? # Y- of reconstruction volume\n", "UTF-8"
                    )
                )
            if fileParameter["end_voxel_3"] is not None:
                fidparfile.write(
                    bytes(
                        "END_VOXEL_3 =   %4d # Z- of reconstruction volume\n"
                        % fileParameter["end_voxel_3"],
                        "UTF-8",
                    )
                )
            else:
                fidparfile.write(
                    bytes(
                        "END_VOXEL_3 =   ??? # Z- of reconstruction volume\n", "UTF-8"
                    )
                )
            fidparfile.write(
                bytes(
                    "OVERSAMPLING_FACTOR = %d # 0 = Linear, 1 = Nearest pixel\n"
                    % fileParameter["oversampling_factor"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes(
                    "ANGLE_OFFSET = %f # Reconstruction rotation offset angle in degrees\n"
                    % recons_param["FT"]["ANGLE_OFFSET_VALUE"],
                    "UTF-8",
                )
            )

            # adding extra features for PyHST
            fidparfile.write(bytes("\n# Parameters extra features PyHST\n", "UTF-8"))
            fidparfile.write(
                bytes(
                    "DO_CCD_FILTER = %s # CCD filter (spikes)\n"
                    % fileParameter["do_ccd_filter"],
                    "UTF-8",
                )
            )
            fidparfile.write(
                bytes('CCD_FILTER = "%s"\n' % fileParameter["ccd_filter"], "UTF-8")
            )
            assert isinstance(fileParameter["ccd_filter_para"], dict)
            fidparfile.write(
                bytes(
                    "CCD_FILTER_PARA = %s\n" % fileParameter["ccd_filter_para"], "UTF-8"
                )
            )
            fidparfile.write(
                bytes(
                    "DO_SINO_FILTER = %s # Sinogram filter (rings)\n"
                    % fileParameter["do_sino_filter"],
                    "UTF-8",
                )
            )
            fidparfile.write(bytes('SINO_FILTER = "%s"\n' % "SINO_Filter", "UTF-8"))
            if num_image_1 is not None:
                fidparfile.write(
                    bytes("ar = Numeric.ones(%d,'f')\n" % num_image_1, "UTF-8")
                )
            fidparfile.write(bytes("ar[0]=0.0\n", "UTF-8"))

            if fileParameter["correct_rings_nb"] > 0:
                fidparfile.write(
                    bytes(
                        "ar[2:%d]=0.0\n" % 2 * (fileParameter["correct_rings_nb"] + 1),
                        "UTF-8",
                    )
                )

            fidparfile.write(bytes('SINO_FILTER_PARA = {"FILTER": ar }\n', "UTF-8"))

            # correct_file should include one correction per projection.
            # for now no relation with axis center of rotation

            # if correct_file is None:
            do_axis_correction = recons_param["FTAXIS"]["DO_AXIS_CORRECTION"]
            # WARNING: For now axis correction is not handled
            do_axis_correction = False
            if do_axis_correction in (True, "YES", 1, "1"):
                do_axis_correction = "YES"
            else:
                do_axis_correction = "NO"
            fidparfile.write(
                bytes(
                    "DO_AXIS_CORRECTION = {} # Axis correction\n"
                    "".format(do_axis_correction),
                    "UTF-8",
                )
            )
            correct_file = "correct.txt"
            if do_axis_correction and os.path.exists(
                os.path.join(scan.path, correct_file)
            ):
                fidparfile.write(
                    bytes("AXIS_CORRECTION_FILE = %s\n" % correct_file, "UTF-8")
                )

            fidparfile.write(
                bytes(
                    "OPTIONS= { 'padding':'%s', 'axis_to_the_center':'%s', "
                    "'avoidhalftomo':'%s'} # Padding and position axis\n"
                    ""
                    % (
                        fileParameter["padding"],
                        fileParameter["axis_to_the_center"],
                        fileParameter["avoidhalftomo"],
                    ),
                    "UTF-8",
                )
            )
            # fidparfile.write('NSLICESATONCE=200  # limiting the number of slices at once to not crash the memory \n')

            if paganin is not PaganinMode.off:
                fidparfile.write(
                    bytes("\n# Parameters for Paganin reconstruction\n", "UTF-8")
                )
                fidparfile.write(
                    bytes(
                        "# delta over beta ratio for Paganin phase retrieval = %1.2f \n"
                        % PAG["DB"],
                        "UTF-8",
                    )
                )
                fidparfile.write(bytes("DO_PAGANIN = 1 \n", "UTF-8"))
                fidparfile.write(
                    bytes("PAGANIN_Lmicron = %f \n" % float(PAG["DB"]), "UTF-8")
                )

                if fileParameter["start_voxel_3"] == fileParameter["end_voxel_3"]:
                    # disp('file for reconstruction of a single slice, I reduce the part of the picture taken into account for the phase retrieval for faster processing')
                    fidparfile.write(bytes("PAGANIN_MARGE = 50 \n", "UTF-8"))
                else:
                    # disp('preparation for a full reconstruction of the volume, the whole picture will be used for the phase retrieval process to ensure final quality and accuracy')
                    fidparfile.write(bytes("PAGANIN_MARGE = 200 \n", "UTF-8"))

                fidparfile.write(bytes("DO_OUTPUT_PAGANIN = 0 \n", "UTF-8"))
                fidparfile.write(bytes("OUTPUT_PAGANIN_FILE = paga_cufft \n", "UTF-8"))
                fidparfile.write(bytes("PAGANIN_TRY_CUFFT = 1 \n", "UTF-8"))
                fidparfile.write(bytes("PAGANIN_TRY_FFTW = 1 \n", "UTF-8"))

                if fileParameter["do_unsharp"] == 1:
                    fidparfile.write(
                        bytes(
                            "\n# Parameters for unsharp masking on the radiographs\n",
                            "UTF-8",
                        )
                    )
                    fidparfile.write(bytes("UNSHARP_LOG = 1 \n", "UTF-8"))
                    fidparfile.write(
                        bytes(
                            "PUS = %f \n" % recons_param["PAGANIN"]["UNSHARP_SIGMA"],
                            "UTF-8",
                        )
                    )
                    fidparfile.write(
                        bytes(
                            "PUC = %f \n" % recons_param["PAGANIN"]["UNSHARP_COEFF"],
                            "UTF-8",
                        )
                    )

            if paganin is PaganinMode.multi:
                fidparfile.write(bytes("\n# Multi_paganin\n", "UTF-8"))
                fidparfile.write(
                    bytes(
                        "# 2nd delta over beta ratio for multi Paganin phase retrieval = %1.2f \n"
                        % PAG["DB2"],
                        "UTF-8",
                    )
                )
                fidparfile.write(bytes("MULTI_PAGANIN_PARS={}\n", "UTF-8"))
                # TODO: check the paganin values
                # I don't know what lmicron is related to
                fidparfile.write(
                    bytes(
                        'MULTI_PAGANIN_PARS["BONE_Lmicron"]= %f \n' % PAG["DB2"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        'MULTI_PAGANIN_PARS["THRESHOLD"]= %1.3f\n' % PAG["THRESHOLD"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        'MULTI_PAGANIN_PARS["DILATE"]= %1.0f\n' % PAG["DILATE"], "UTF-8"
                    )
                )
                fidparfile.write(
                    bytes(
                        'MULTI_PAGANIN_PARS["MEDIANR"]= %1.0f\n'
                        % recons_param["PAGANIN"]["MEDIANR"],
                        "UTF-8",
                    )
                )

                if fileParameter["multipag_keep_bone"] == 1:
                    fidparfile.write(
                        bytes(
                            'MULTI_PAGANIN_PARS["PAGBONE_FILE"] = "%s"\n'
                            % fileParameter["output_file_bone"],
                            "UTF-8",
                        )
                    )
                if fileParameter["multipag_keep_mask"] == 1:
                    fidparfile.write(
                        bytes(
                            'MULTI_PAGANIN_PARS["MASKBONE_FILE"] = "%s"\n'
                            % fileParameter["output_file_mask"],
                            "UTF-8",
                        )
                    )
                if fileParameter["multipag_keep_abs"] == 1:
                    fidparfile.write(
                        bytes(
                            'MULTI_PAGANIN_PARS["ABSBONE_FILE"] = "%s"\n'
                            % fileParameter["output_file_abs"],
                            "UTF-8",
                        )
                    )
                if fileParameter["multipag_keep_corr"] == 1:
                    fidparfile.write(
                        bytes(
                            'MULTI_PAGANIN_PARS["CORRBONE_FILE"] = "%s"\n'
                            % fileParameter["output_file_corr"],
                            "UTF-8",
                        )
                    )
                if fileParameter["multipag_keep_soft"] == 1:
                    fidparfile.write(
                        bytes(
                            'MULTI_PAGANIN_PARS["CORRECTEDVOL_FILE"] = "%s"\n'
                            % fileParameter["output_file_soft"],
                            "UTF-8",
                        )
                    )

            if fileParameter["half_acquisition"] == 1:
                fidparfile.write(bytes("PENTEZONE=300\n", "UTF-8"))

            if FT["ZEROOFFMASK"] == 0:
                fidparfile.write(bytes("ZEROOFFMASK=0\n", "UTF-8"))

            ########################## addings for version 2 of Pyhst ################################

            fidparfile.write(
                bytes("TRYEDFCONSTANTHEADER=%d \n" % (1 - FT["FIXHD"]), "UTF-8")
            )

            ##########################################################################################

            if "zeroclipvalue" not in fileParameter:
                fidparfile.write(
                    bytes(
                        "ZEROCLIPVALUE = %g # Minimum value of radiographs after flat / before log\n"
                        % fileParameter["zeroclipvalue"],
                        "UTF-8",
                    )
                )

            if "oneclipvalue" not in fileParameter:
                fidparfile.write(
                    bytes(
                        "ONECLIPVALUE = %g # Maximum value of radiographs after flat / before log\n"
                        % fileParameter["oneclipvalue"],
                        "UTF-8",
                    )
                )

            if fileParameter["do_projection_median"] == "YES":
                fidparfile.write(
                    bytes(
                        "DO_PROJECTION_MEDIAN = %s # Calculate median of all projections\n"
                        % fileParameter["do_projection_median"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        'PROJECTION_MEDIAN_FILENAME = "%s" # Name output file median calculation\n'
                        % fileParameter["projection_median_filename"],
                        "UTF-8",
                    )
                )

            if fileParameter["do_projection_mean"] == "YES":
                fidparfile.write(
                    bytes(
                        "DO_PROJECTION_MEAN = %s # Calculate median of all projections\n"
                        % fileParameter["do_projection_mean"],
                        "UTF-8",
                    )
                )
                fidparfile.write(
                    bytes(
                        'PROJECTION_MEDIAN_FILENAME = "%s" # Name output file mean calculation\n'
                        % fileParameter["projection_mean_filename"],
                        "UTF-8",
                    )
                )

            fidparfile.write(
                bytes("\n# Parameters defining output file / format\n", "UTF-8")
            )
            fidparfile.write(
                bytes("OUTPUT_FILE = %s\n" % fileParameter["output_file"], "UTF-8")
            )

            fidparfile.write(bytes("\n# Reconstruction program options\n", "UTF-8"))
            # old legacy, not needed by pyhst2
            # fidparfile.write(
            #     bytes('DISPLAY_GRAPHICS = %s # No images\n' % fileParameter['display_graphics'],
            #           'UTF-8'))

            if fileParameter["ft_jp2"] == 1:
                fidparfile.write(bytes("\n#JP2000\n", "UTF-8"))
                fidparfile.write(bytes('JP2EDF_DIR = "/tmp/jp2/" \n', "UTF-8"))
                fidparfile.write(bytes("JP2EDF_REMOVE = 1 \n", "UTF-8"))

            ##########CONICAL AND FAN BEAM RECONSTRUCTION############################################

            if BG["TYPE"] == "c":
                fidparfile.write(
                    bytes("\n# Conical and Fan beam reconstruction\n", "UTF-8")
                )
                fidparfile.write(bytes("CONICITY = 1\n", "UTF-8"))
                fidparfile.write(bytes("CONICITY_FAN = 0\n", "UTF-8"))
                fidparfile.write(
                    bytes("SOURCE_X = %f\n" % fileParameter["source_x"], "UTF-8")
                )
                fidparfile.write(
                    bytes("SOURCE_Y = %f\n" % fileParameter["source_y"], "UTF-8")
                )
                fidparfile.write(
                    bytes(
                        "SOURCE_DISTANCE = %f\n" % fileParameter["source_distance"],
                        "UTF-8",
                    )
                )

            if BG["TYPE"] == "f":
                fidparfile.write(
                    bytes("\n# Conical and Fan beam reconstruction\n", "UTF-8")
                )
                fidparfile.write(bytes("CONICITY = 0\n", "UTF-8"))
                fidparfile.write(bytes("CONICITY_FAN = 1\n", "UTF-8"))
                fidparfile.write(
                    bytes("SOURCE_X = %f\n" % fileParameter["source_x"], "UTF-8")
                )
                fidparfile.write(
                    bytes("SOURCE_Y = %f\n" % fileParameter["source_y"], "UTF-8")
                )
                fidparfile.write(
                    bytes(
                        "SOURCE_DISTANCE = %f\n" % fileParameter["source_distance"],
                        "UTF-8",
                    )
                )

        return par_file_path

    @staticmethod
    def make_rec_file(rec_file, par_file, recons_params):
        """
        Function creating the .rec file

        :param rec_file: name of the output file
        :type: str
        :param recons_params: reconstruction parameters as dict
        :type: dict
        """

        def loadFileInfoFrmPar(filePath):
            info = "Reading %s to get reconstruction parameters"
            logger.info(info)
            ofile = open(filePath, "r")
            lines = ofile.readlines()
            outputFile = None
            filePrefix = None
            for line in lines:
                line_str = str(line)
                if "OUTPUT_FILE" in line_str:
                    splitLine = line_str.replace(" ", "").split("=")
                    outputFile = splitLine[1].rstrip("\n")
                    logger.info("Found OUTPUT_FILE: %s" % outputFile)

                if "FILE_PREFIX" in line_str:
                    splitLine = line_str.replace(" ", "").split("=")
                    filePrefix = splitLine[1].rstrip("\n")
                    logger.info("Found FILE_PREFIX: %s" % filePrefix)
            ofile.close()
            return outputFile, filePrefix

        test = recons_params["FT"]

        vol_output = test["VOLOUTFILE"]

        ### Get useful parameter
        if os.path.isfile(par_file):
            recFile, prefix = loadFileInfoFrmPar(par_file)
        else:
            err = "No %s file found. Can't deduce reconstruction parameters." % par_file
            raise ValueError(err)

        if recFile is None or prefix is None:
            raise IOError(
                "Failed to load parameters `OUTPUT_FILE` and/or "
                "`FILE_PREFIX` in the reconstruction parameters file"
            )

        ### Get the prefix for .rec name file
        ### prefix : only name of file without extension
        ### prefix_full : name and directory of file without extension
        basename1 = os.path.basename(prefix)
        logger.info("prefix is %s" % basename1)
        prefix_full = par_file[0:-4]
        logger.info("prefix is %s" % prefix_full)
        if prefix_full[-3:] == "pag":
            prefix = prefix_full[0:-3]
        else:
            prefix = prefix_full

        # Get parameter for writing the script
        # utilsMakeOAR = recons_params['PYHSTEXE']['MAKE_OAR_FILE']
        # For now hard coded
        utilsMakeOAR = settings.MAKE_OAR_PYST2_PATH

        logger.info(("Processing the file %s %s" % (prefix_full, prefix)))

        # Writing the script file
        with open(rec_file, "wb") as fidshell:
            fidshell.write(bytes("#!/bin/bash\n", "UTF-8"))
            fidshell.write(
                bytes('GET_OS="/csadmin/common/scripts/get_os.share"\n', "UTF-8")
            )
            fidshell.write(bytes("os=`$GET_OS`\n", "UTF-8"))
            fidshell.write(
                bytes(
                    "export PYHST="
                    + _getPyHSTDir()
                    + recons_params["PYHSTEXE"]["EXE"]
                    + "\n",
                    "UTF-8",
                )
            )
            fidshell.write(bytes('case "${os}" in\n', "UTF-8"))
            fidshell.write(bytes("    centos*)\n", "UTF-8"))
            if vol_output == 0:
                fidshell.write(
                    bytes("      mkdir " + os.path.dirname(recFile) + " \n", "UTF-8")
                )
                fidshell.write(
                    bytes(
                        "      chmod 775 " + os.path.dirname(recFile) + " \n", "UTF-8"
                    )
                )

            fidshell.write(
                bytes(
                    "        "
                    + str(utilsMakeOAR)
                    + " "
                    + os.path.basename(prefix_full)
                    + " 1\n",
                    "UTF-8",
                )
            )

            fidshell.write(
                bytes(
                    "        oarsub -q gpu  -S ./"
                    + os.path.basename(prefix_full)
                    + ".oar\n",
                    "UTF-8",
                )
            )

            fidshell.write(bytes("        ;;\n", "UTF-8"))
            fidshell.write(bytes("    debian*)\n", "UTF-8"))
            if vol_output == 0:
                fidshell.write(
                    bytes("      mkdir " + os.path.dirname(recFile) + " \n", "UTF-8")
                )

            fidshell.write(
                bytes(
                    "        "
                    + str(utilsMakeOAR)
                    + " "
                    + os.path.basename(prefix_full)
                    + " 1\n",
                    "UTF-8",
                )
            )
            fidshell.write(
                bytes(
                    "        oarsub -q gpu -S ./"
                    + os.path.basename(prefix_full)
                    + ".oar\n",
                    "UTF-8",
                )
            )

            fidshell.write(bytes("        ;;\n", "UTF-8"))
            fidshell.write(bytes("    *)\n", "UTF-8"))
            fidshell.write(bytes("esac\n", "UTF-8"))

        os.chmod(rec_file, 0o777)
        return

    @staticmethod
    def _loadDefaultPar(
        recons_param,
        out_base_file_name,
        direc,
        image_pixel_size,
        corr,
        nvue,
        refon,
        num_image_1,
        num_image_2,
        slice,
        convert_lbsram,
        cor,
        scan_range,
    ):
        assert type(recons_param) is dict
        assert out_base_file_name is not None
        # TODO: default params should not be necessary and should be removed
        defaultParams = {}
        FT = recons_param["FT"]
        PG = PAG = recons_param["PAGANIN"]
        BG = recons_param["BEAMGEO"]
        paganin = PaganinMode.from_value(PAG["MODE"])

        defaultParams["half_acquisition"] = FT["HALF_ACQ"]

        if PAG["UNSHARP_SIGMA"] != 0:
            defaultParams["do_unsharp"] = 1
        else:
            defaultParams["do_unsharp"] = 0

        defaultParams["correct_spikes_threshold_default"] = 0.04
        defaultParams["status"] = 1
        defaultParams["parfilename"] = out_base_file_name + ".par"

        defaultParams["num_first_image"] = 0
        defaultParams["num_last_image"] = recons_param["FT"]["NUM_LAST_IMAGE"] - 1
        defaultParams["angle_offset"] = 0

        defaultParams["number_length_varies"] = "NO"
        defaultParams["file_postfix"] = ".edf"
        defaultParams["image_pixel_size_1"] = defaultParams[
            "image_pixel_size_2"
        ] = image_pixel_size
        defaultParams["file_interval"] = 1
        defaultParams["subtract_background"] = defaultParams[
            "correct_flatfield"
        ] = corr  # = take_logarithm
        defaultParams["take_logarithm"] = "YES"
        defaultParams["correct"] = corr == "YES"
        defaultParams["background_file"] = os.path.join(direc, "dark.edf")
        defaultParams["ff_prefix"] = os.path.join(direc, "refHST")
        defaultParams["flatfield_file"] = defaultParams["ff_prefix"] + "0000.edf"
        defaultParams["ff_num_first_image"] = 0
        defaultParams["ff_num_last_image"] = nvue
        defaultParams["ff_file_interval"] = refon

        if defaultParams["ff_file_interval"] == 0:
            defaultParams["flatfield_changing"] = "NO"
        else:
            defaultParams["flatfield_changing"] = "YES"

        defaultParams["doubleffcorrection"] = []
        defaultParams["zeroclipvalue"] = 0.001
        defaultParams["oneclipvalue"] = []
        defaultParams["multiframe"] = 0
        defaultParams["angle_between_projections"] = (
            scan_range / nvue if nvue != 0 else 0
        )
        defaultParams["horaxis"] = 0
        defaultParams["output_sinograms"] = "NO"
        defaultParams["output_reconstruction"] = "YES"
        defaultParams["start_voxel_1"] = defaultParams["start_voxel_2"] = 1
        if slice is None:
            defaultParams["start_voxel_3"] = 1
        else:
            defaultParams["start_voxel_3"] = (
                slice + 1
            )  # VOXEL start at 1 in pyhst, 0 in tomwer
        defaultParams["end_voxel_1"] = defaultParams["end_voxel_2"] = num_image_1
        if recons_param["FT"]["HALF_ACQ"]:
            if cor is None:
                cor = num_image_1 // 2
            defaultParams["end_voxel_1"] = defaultParams[
                "end_voxel_2"
            ] = num_image_1 + abs(2.0 * cor)
        if slice is None:
            defaultParams["end_voxel_3"] = num_image_2
        else:
            defaultParams["end_voxel_3"] = (
                slice + 1
            )  # VOXEL start at 1 in pyhst, 0 in tomwer
        defaultParams["oversampling_factor"] = 4

        defaultParams["ccd_filter"] = "CCD_Filter"

        # correction rings
        if FT["RINGSCORRECTION"]:
            defaultParams["do_sino_filter"] = "YES"
        else:
            defaultParams["do_sino_filter"] = "NO"

        defaultParams[
            "correct_rings_nb"
        ] = 2  # number of elements (|f|>0) that will not be filtered from sinograms
        defaultParams["sino_filter"] = "SINO_Filter"

        # padding at edges 'E' edge values or '0'
        defaultParams["padding"] = "E"
        # move axis to the center 'Y' or not 'N'
        defaultParams["axis_to_the_center"] = "Y"
        defaultParams["do_projection_median"] = "NO"
        defaultParams["projection_median_filename"] = "median.edf"
        defaultParams["do_projection_mean"] = "NO"
        defaultParams["projection_mean_filename"] = "mean.edf"
        # defaultParams['display_graphics'] = 'NO'

        if (BG["TYPE"] == "c") or (BG["TYPE"] == "f"):
            defaultParams["source_x"] = BG["SX"]
            defaultParams["source_y"] = BG["SY"]
            defaultParams["source_distance"] = BG["DIST"]
        else:
            defaultParams["source_x"] = 0
            defaultParams["source_y"] = 0
            defaultParams["source_distance"] = 0

        valJP2 = FT["JP2"] if "JP2" in FT else 0
        defaultParams["ft_jp2"] = valJP2

        if FT["VOLOUTFILE"] == 1:
            output_file_ext = ".vol"
        else:
            output_file_ext = ".edf"
        if paganin is PaganinMode.off:
            defaultParams["output_file"] = os.path.join(direc, out_base_file_name)
        elif paganin in (PaganinMode.on, PaganinMode.both):
            # TODO: those are missing, I don't know how to compute them
            defaultParams["pag_length"] = 500.0
            defaultParams["output_file"] = os.path.join(direc, out_base_file_name)
        elif paganin is PaganinMode.multi:
            defaultParams["output_file"] = os.path.join(direc, out_base_file_name)
            defaultParams["ft_pag_db2"] = PG["DB2"]
            defaultParams["ft_pag_threshold"] = PG["THRESHOLD"]
            defaultParams["ft_pag_dilate"] = PG["DILATE"]
            defaultParams["ft_pag_medianr"] = PG["MEDIANR"]
            defaultParams["multipag_keep_bone"] = PG["MKEEP_BONE"]
            defaultParams["multipag_keep_soft"] = PG["MKEEP_SOFT"]
            defaultParams["multipag_keep_abs"] = PG["MKEEP_ABS"]
            defaultParams["multipag_keep_corr"] = PG["MKEEP_CORR"]
            defaultParams["multipag_keep_mask"] = PG["MKEEP_MASK"]
            # probleme a regler
            defaultParams["multipag_extra_files"] = (
                defaultParams["multipag_keep_bone"] == 1
                or defaultParams["multipag_keep_soft"] == 1
                or defaultParams["multipag_keep_mask"] == 1
                or defaultParams["multipag_keep_abs"] == 1
                or defaultParams["multipag_keep_corr"] == 1
            )
            defaultParams["output_file_bone"] = os.path.join(
                direc, out_base_file_name + "multipag_bone" + output_file_ext
            )
            defaultParams["output_file_soft"] = os.path.join(
                direc, out_base_file_name + "multipag_soft" + output_file_ext
            )
            defaultParams["output_file_abs"] = os.path.join(
                direc, out_base_file_name + "multipag_abs" + output_file_ext
            )
            defaultParams["output_file_corr"] = os.path.join(
                direc, out_base_file_name + "multipag_corr" + output_file_ext
            )
            defaultParams["output_file_mask"] = os.path.join(
                direc, out_base_file_name + "multipag_mask" + output_file_ext
            )
        else:
            raise ValueError("Paganin mode not recognized")

        # if necessary convert lbsram path to destination path
        keys_defining_path = (
            "output_file",
            "background_file",
            "ff_prefix",
            "flatfield_file",
            "output_file_bone",
            "output_file_soft",
            "output_file_abs",
            "output_file_corr",
            "output_file_mask",
        )
        if convert_lbsram:
            for key in keys_defining_path:
                if key in defaultParams:
                    value = defaultParams[key]
                    defaultParams[key] = PyHSTCaller._remove_lbs_head(dirin=value)

        return defaultParams

    @staticmethod
    def _remove_lbs_head(dirin):
        if dirin.startswith(get_lbsram_path()):
            return dirin.replace(get_lbsram_path(), get_dest_path(), 1)
        else:
            return dirin



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Unit test for the scan defined at the hdf5 format"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/11/2021"


import asyncio
from tomwer.core.scan.futurescan import FutureTomwerScan
from tomwer.core.utils.scanutils import MockHDF5
import os


def test_simple_future_scan(tmpdir):
    """Simple test of the FutureTomwerScan API"""
    scan = MockHDF5(
        scan_path=os.path.join(tmpdir, "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        create_ini_dark=False,
        create_ini_ref=False,
        dim=10,
    ).scan
    future = asyncio.Future()
    future.set_result(None)

    future_scan = FutureTomwerScan(
        scan=scan,
        futures=[
            future,
        ],
    )

    future_scan.results()
    assert future_scan.exceptions() is None
    assert future_scan.cancelled() is False



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""convert from NXTomo to hdf5 compliant for pyhst"""

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/04/2020"


import h5py
from collections import namedtuple
import os
import silx.io.utils
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomoscan.io import HDF5File


_pyhst_hdf5_info = namedtuple("_pyhst_hdf5_info", ["dark", "flats", "projections"])


_pyhst_hdf5_dark = namedtuple("_pyhst_hdf5_dark", ["file", "dataset_path"])

_pyhst_hdf5_flats = namedtuple(
    "_pyhst_hdf5_flats", ["files", "dataset_path", "intervals"]
)

_pyhst_hdf5_projections = namedtuple(
    "_pyhst_hdf5_projections", ["file", "dataset_path"]
)


def _create_file_with_virtual_dataset(
    vsource, pro_indxs, datasource_shape, mydtype, filename
):
    layout = h5py.VirtualLayout(
        shape=tuple([len(pro_indxs)] + list(datasource_shape)[1:]), dtype=mydtype
    )
    for n in range(len(pro_indxs)):
        layout[n] = vsource[pro_indxs[n]]
    with HDF5File(filename, "w") as f:
        f.create_virtual_dataset("data", layout=layout, fillvalue=-5)


def _copy_dataset(data, output_file):
    with HDF5File(output_file, "w") as f:
        f.create_dataset(name="data", data=data)


def convert_nx_to_pyhst_hdf5(scan: HDF5TomoScan) -> tuple:
    """
    Convert an HDF5TomoScan (NXTomo compliant) to 3 HDF5 compliant with pyhst:
    one for dark, one for flats and one for projections

    :param HDF5TomoScan scan: scan to convert
    :return: _pyhst_hdf5_info
            note: we create one file per flat.
    :rtype: tuple
    """
    assert isinstance(scan, HDF5TomoScan)
    output_dir = scan.path
    input_master_file = scan.master_file
    entry_name = scan.entry

    entry = HDF5File(input_master_file, "r", swmr=True)[entry_name]

    pro_indxs = []
    ff_indxs = []
    ff_poss = []
    bck_indxs = []

    keys = entry["instrument/detector"]["image_key"]

    old_ff_pos = -1
    ffblock = None
    ipro = 0
    for i, k in enumerate(list(keys)):
        if k == 2:
            bck_indxs.append(i)
        if k == 0:
            pro_indxs.append(i)
            ipro += 1
        if k == 1:
            if ipro == old_ff_pos:
                ffblock.append(i)
            else:
                ffblock = []
                ff_indxs.append(ffblock)
                ff_poss.append(ipro)
                old_ff_pos = ipro

    datasource_shape = entry["instrument/detector/data"].shape
    mydtype = entry["instrument/detector/data"].dtype

    vsource = h5py.VirtualSource(
        input_master_file,
        entry_name + "/instrument/detector/data",
        shape=datasource_shape,
    )

    # manage projections
    file_basename = ".".join(os.path.splitext(os.path.basename(scan.master_file))[:-1])
    proj_filename = "pyhst2_projs_{entry}_{basename}.h5".format(
        entry=str(scan.entry), basename=file_basename
    )
    proj_filename = os.path.join(output_dir, proj_filename)
    _create_file_with_virtual_dataset(
        vsource, pro_indxs, datasource_shape, mydtype, proj_filename
    )
    proj_ds_name = '"data"'

    projections_info = _pyhst_hdf5_projections(proj_filename, proj_ds_name)

    # manage dark
    def get_dark():
        if not os.path.exists(scan.process_file):
            return {}

        return DarkRefs.get_darks_frm_process_file(
            process_file=scan.process_file, entry=scan.entry, as_url=True
        )

    dark_url = get_dark()
    if len(dark_url) > 0:
        url = dark_url[list(dark_url.keys())[0]]
        dark_filename = "pyhst2_dark_{entry}_{basename}.h5".format(
            entry=str(scan.entry), basename=file_basename
        )
        dark_filename = os.path.join(output_dir, dark_filename)
        _copy_dataset(data=silx.io.utils.get_data(url), output_file=dark_filename)

        dark_ds_name = '"data"'
        dark_info = _pyhst_hdf5_dark(dark_filename, dark_ds_name)
    else:
        dark_info = None

    def get_flats():
        if not os.path.exists(scan.process_file):
            return {}

        return DarkRefs.get_flats_frm_process_file(
            process_file=scan.process_file, entry=scan.entry, as_url=True
        )

    # manage flats
    def adapt_flat_index_to_pyhst(urls):
        """as this is an hdf5 acquisition we now have the index of the
        acquisition but in a different referential."""
        assert len(urls) > 0
        res = {}
        if 0 in urls:
            res[0] = urls[0]
        elif 1 in urls:
            res[0] = urls[1]
        if len(urls) > 1:
            keys = list(urls.keys())
            # now consider the first projection to be 0
            last_index = keys[1] - list(scan.projections.keys())[0]
            res[last_index] = urls[keys[-1]]
        return res

    flats_url = get_flats()
    if len(flats_url):
        flats_url = adapt_flat_index_to_pyhst(flats_url)

        flats_files = []
        flats_intervals = []
        for flat_index, url in flats_url.items():
            flats_intervals.append(flat_index)
            fname = "pyhst2_flat_{:04d}_{entry}_{basename}.h5".format(
                flat_index, entry=str(scan.entry), basename=file_basename
            )
            fname = os.path.join(output_dir, fname)
            flats_files.append(fname)
            flat_data = silx.io.utils.get_data(url)
            if flat_data.ndim == 2:
                flat_data = flat_data.reshape(1, flat_data.shape[0], flat_data.shape[1])
            _copy_dataset(data=flat_data, output_file=fname)
        flats_ds_name = '"data"'
        flats_info = _pyhst_hdf5_flats(flats_files, flats_ds_name, flats_intervals)
    else:
        flats_info = None
    return _pyhst_hdf5_info(
        dark=dark_info, flats=flats_info, projections=projections_info
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/02/2017"

from silx.gui import qt
import os


class H5FileDialog(qt.QMessageBox):
    """Class defining the dialog to ask the user an h5 file to be used"""

    defaultH5File = os.path.expanduser("~") + "/.octave/mytomodefaults.h5"

    def __init__(self, parent=None):
        qt.QMessageBox.__init__(self, parent)
        self.setWindowTitle(".h5 file for reconstruction")

        self.h5File = None

        text = "No .h5 found, if you want to process a reconstruction "
        text += "you must give one."
        self.layout().addWidget(qt.QLabel(text), 1, 0)

        if H5FileDialog.hasDefaultH5File():
            buttonText = "Take the default octave file at " + self.defaultH5File
            defaultOctaveFileButton = qt.QPushButton(buttonText)
            defaultOctaveFileButton.setAutoDefault(True)
            defaultOctaveFileButton.clicked.connect(self.__selectDefaultH5File)
            self.layout().addWidget(defaultOctaveFileButton, 1, 0)

        selectFileButtonButton = qt.QPushButton("Select a file")
        self.layout().addWidget(selectFileButtonButton, 2, 0)
        selectFileButtonButton.setAutoDefault(True)

        # rearange the usual QMesageBox buttons
        self.cancelButton = self.addButton(qt.QMessageBox.Cancel)
        self.layout().removeWidget(self.cancelButton)
        self.layout().addWidget(self.cancelButton, 3, 0)
        selectFileButtonButton.clicked.connect(self.__selectFileFromDlg)

    def __selectDefaultH5File(self):
        """
        Set the default h5 file as the selected one
        """
        self.h5File = self.defaultH5File
        self.accept()

    def __selectFileFromDlg(self):
        """
        Launch a QFileDialog to ask the user the path to the h5 file
        """
        self.h5File = H5FileDialog.askForH5File()
        self.accept()

    def exec_(self):
        # If no default h5 file only ask for one.
        if H5FileDialog.hasDefaultH5File():
            qt.QMessageBox.exec_(self)
        else:
            H5FileDialog.askForH5File()

    @staticmethod
    def hasDefaultH5File():
        return os.path.isfile(H5FileDialog.defaultH5File)

    @staticmethod
    def askForH5File(save):
        dialog = qt.QFileDialog()
        title = "Select an h5 file to %s the reconstruction parameters." "" % (
            "save" if save is True else "load"
        )
        dialog.setWindowTitle(title)
        dialog.setModal(1)
        dialog.setNameFilters(["HDF5 file *.h5 *.hdf5"])

        if not dialog.exec_():
            return None

        qapp = qt.QApplication.instance()
        qapp.processEvents()
        return dialog.selectedFiles()[0]
